# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl.power.rte import ConvectiveCooling


conv_cool_instances = [
    ConvectiveCooling(
        altitude=np.array([100.0]),
        cable_azimuth=np.array([2]),
        ambient_temperature=np.array([25.0]),
        wind_speed=np.array([10.0]),
        wind_azimuth=np.array([11.0]),
        outer_diameter=np.array([0.01]),
        solar_absorptivity=np.array([0.5]),
    ),
    ConvectiveCooling(
        altitude=100.0,
        cable_azimuth=2,
        ambient_temperature=25.0,
        wind_speed=10.0,
        wind_azimuth=11.0,
        outer_diameter=0.01,
        solar_absorptivity=0.5,
    ),
]


@pytest.mark.parametrize(
    "convective_cooling",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_convective_cooling_value(convective_cooling):
    temperature = np.array([30.0])
    expected_value = 9.50218

    result = convective_cooling.value(temperature)

    np.testing.assert_allclose(result, expected_value, rtol=1e-5)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone

from zoneinfo import ZoneInfo
import os.path
import numpy as np
import pandas as pd
from thermohl.solver import rte
from thermohl.solver.entities import VariableType, HeatEquationType, TemperatureType


def get_cable_data(cable_name: str) -> dict:
    """Get cable/conductor data from file."""
    f = os.path.join("test", "functional_test", "cable_catalog.csv")
    df = pd.read_csv(f)
    if cable_name in df["conductor"].values:
        return df[df["conductor"] == cable_name].to_dict(orient="records")[0]
    else:
        raise ValueError(f"Conductor {cable_name} not found in file {f}.")


def get_scenarios(scenario_file_name: str):
    f = os.path.join("test", "functional_test", scenario_file_name)
    df = pd.read_csv(f, sep=";")
    records = df.astype(object).where(pd.notna(df), None).to_dict(orient="records")
    return records


def scn2dict(scn: dict) -> dict:
    """Convert scenario to thermohl input."""
    dic = get_cable_data(scn["Conducteur"])

    dic["latitude"] = scn["Lat"]
    dic["longitude"] = scn["Lon"]
    dic["altitude"] = scn["Altitude"]
    dic["cable_azimuth"] = scn["Azimut_portee"]
    dic["ambient_temperature"] = scn["T_amb"]
    dic["wind_speed"] = scn["V"]
    dic["wind_azimuth"] = scn["Azimut_V"]
    dic["albedo"] = scn["Albedo"]
    dic["measured_solar_irradiance"] = scn["QG_mesure"]
    dic["nebulosity"] = scn["Neb"]
    dic["solar_absorptivity"] = 0.9
    dic["emissivity"] = 0.8

    datetime_paris = datetime.strptime(
        f'{scn["Date"]} {scn["Heure"]}', "%d/%m/%Y %H:%M"
    ).replace(tzinfo=ZoneInfo("Europe/Paris"))
    datetime_utc = datetime_paris.astimezone(timezone.utc)
    dic["datetime_utc"] = datetime_utc
    # utile uniquement pour le test test_steady_temperature
    if "Intensite" in scn:
        dic[VariableType.TRANSIT.value] = scn["Intensite"]

    return dic


def test_steady_temperature():
    for scenario in get_scenarios("scenarios_steady.csv"):
        solver = rte(
            scn2dict(scenario),
            heat_equation=HeatEquationType.THREE_TEMPERATURES_LEGACY,
        )
        result = solver.steady_temperature()
        assert np.allclose(
            result[TemperatureType.SURFACE], scenario["T_surf"], atol=0.05
        )
        assert np.allclose(
            result[TemperatureType.AVERAGE], scenario["T_moy"], atol=0.05
        )
        assert np.allclose(result[TemperatureType.CORE], scenario["T_coeur"], atol=0.05)


def test_steady_ampacity():
    for scenario in get_scenarios("scenarios_steady.csv"):
        solver = rte(
            scn2dict(scenario),
            heat_equation=HeatEquationType.THREE_TEMPERATURES_LEGACY,
        )
        result = solver.steady_intensity(max_conductor_temperature=scenario["T_conf"])
        assert np.allclose(
            result[VariableType.TRANSIT], scenario["Ampacite"], atol=0.05
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Models to compute equilibrium temperature or max intensity in a conductor."""

from typing import Dict, Any, Optional, Union, Type

from thermohl.power import cigre as cigrep
from thermohl.power import rte as rtep
from thermohl.power import ieee as ieeep
from thermohl.power import olla as ollap

from thermohl.solver.solver import Solver
from thermohl.solver.parameters import Parameters
from thermohl.solver.entities import (
    HeatEquationType,
    PowerType as PowerType,
    ModelType,
    VariableType as VariableType,
)
from thermohl.solver.slv1t import Solver1T
from thermohl.solver.slv3t import Solver3T
from thermohl.solver.slv3t_legacy import Solver3TL

concreteSolverType = Union[Type[Solver1T], Type[Solver3T], Type[Solver3TL]]


def default_values() -> Dict[str, Any]:
    return Parameters().__dict__


def _factory(
    dic: Optional[Dict[str, Any]] = None,
    heat_equation: HeatEquationType = HeatEquationType.ONE_TEMPERATURE,
    model: ModelType = ModelType.IEEE,
) -> Solver:
    solver = create_solver_from_heat_equation(heat_equation)

    model_modules = {
        ModelType.CIGRE: cigrep,
        ModelType.IEEE: ieeep,
        ModelType.OLLA: ollap,
        ModelType.RTE: rtep,
    }

    if model not in model_modules:
        raise ValueError(f"Unsupported solver model: {model}")

    module = model_modules[model]
    return solver(
        dic,
        module.JouleHeating,
        module.SolarHeating,
        module.ConvectiveCooling,
        module.RadiativeCooling,
    )


def create_solver_from_heat_equation(
    heat_equation: HeatEquationType,
) -> concreteSolverType:
    heat_equations_solvers = {
        HeatEquationType.ONE_TEMPERATURE: Solver1T,
        HeatEquationType.THREE_TEMPERATURES: Solver3T,
        HeatEquationType.THREE_TEMPERATURES_LEGACY: Solver3TL,
    }

    if heat_equation not in heat_equations_solvers:
        raise ValueError(f"Invalid HeatEquation value {heat_equation.value}")

    return heat_equations_solvers[heat_equation]


def __solver_model(
    dic: Optional[Dict[str, Any]] = None,
    heat_equation: HeatEquationType = HeatEquationType.ONE_TEMPERATURE,
    model: ModelType = ModelType.CIGRE,
) -> Solver:
    """Get a Solver using a given model and heat equation.

    Args:
        dic (dict | None): Input values. The default is None.
        heat_equation (HeatEquationType): Input heat equation.
        model (ModelType): Solver model to use.

    """
    return _factory(dic, heat_equation=heat_equation, model=model)


def cigre(
    dic: Optional[Dict[str, Any]] = None,
    heat_equation: HeatEquationType = HeatEquationType.ONE_TEMPERATURE,
) -> Solver:
    """Get a Solver using CIGRE models.

    Args:
        dic (dict | None): Input values. The default is None.
        heat_equation (HeatEquationType): Input heat equation.

    """
    return __solver_model(dic=dic, heat_equation=heat_equation, model=ModelType.CIGRE)


def ieee(
    dic: Optional[Dict[str, Any]] = None,
    heat_equation: HeatEquationType = HeatEquationType.ONE_TEMPERATURE,
) -> Solver:
    """Get a Solver using IEEE models.

    Args:
        dic (dict | None): Input values. The default is None.
        heat_equation (HeatEquationType): Input heat equation.

    """
    return __solver_model(dic, heat_equation=heat_equation, model=ModelType.IEEE)


def olla(
    dic: Optional[Dict[str, Any]] = None,
    heat_equation: HeatEquationType = HeatEquationType.ONE_TEMPERATURE,
) -> Solver:
    """Get a Solver using RTE-olla models.

    Args:
        dic (dict | None): Input values. The default is None.
        heat_equation (HeatEquationType): Input heat equation.

    """
    return __solver_model(dic, heat_equation=heat_equation, model=ModelType.OLLA)


def rte(
    dic: Optional[Dict[str, Any]] = None,
    heat_equation: HeatEquationType = HeatEquationType.ONE_TEMPERATURE,
) -> Solver:
    """Get a Solver using RTE models.

    Args:
        dic (dict | None): Input values. The default is None.
        heat_equation (HeatEquationType): Input heat equation.

    """
    return __solver_model(dic, heat_equation=heat_equation, model=ModelType.RTE)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from functional_test.test_functional_steady import (
    get_scenarios,
    scn2dict,
)
from thermohl.solver import rte
from thermohl.solver.entities import VariableType, HeatEquationType, TemperatureType


def test_transient_temperature():
    time_constant = 600.0
    offsets = np.arange(0.0, 1800.0 + 1, 10)

    for scenario in get_scenarios("scenarios_transient.csv"):
        solver = rte(
            scn2dict(scenario),
            heat_equation=HeatEquationType.THREE_TEMPERATURES_LEGACY,
        )

        # initial steady state
        solver.args[VariableType.TRANSIT.value] = scenario["I_initial"]
        solver.update()
        initial_state = solver.steady_temperature()

        # check initial temperature
        assert np.isclose(
            scenario["T_mean_0"],
            initial_state[TemperatureType.AVERAGE][0],
            atol=0.01,
        )

        # final steady state
        solver.args[VariableType.TRANSIT.value] = scenario["I_final"]
        solver.update()
        final_state = solver.steady_temperature(
            surface_temperature_guess=scenario["T_surf_final"],
            core_temperature_guess=scenario["T_heart_final"],
        )

        # check final temperature
        assert np.isclose(
            scenario["T_mean_final"],
            final_state[TemperatureType.AVERAGE][0],
            atol=0.05,
        )

        # transient temperature
        transient_result = solver.transient_temperature_legacy(
            offset=offsets,
            surface_temperature_0=initial_state[TemperatureType.SURFACE],
            core_temperature_0=initial_state[TemperatureType.CORE],
            time_constant=time_constant,
        )

        # get expected temperatures from scenario at the right offsets
        useful_offsets = np.array([0, 1, 2, 3, 5, 10, 30])  # in minutes
        expected_surface_temperatures = [
            scenario[f"T_surf_{offset}"] for offset in useful_offsets
        ]
        expected_mean_temperatures = [
            scenario[f"T_mean_{offset}"] for offset in useful_offsets
        ]
        expected_core_temperatures = [
            scenario[f"T_heart_{offset}"] for offset in useful_offsets
        ]

        atol = 0.01
        # get indexes of the useful offsets in the thermohl computation of transient temperature.
        useful_indexes = np.searchsorted(
            transient_result[VariableType.TIME], useful_offsets * 60
        )
        assert np.allclose(
            transient_result[TemperatureType.SURFACE][useful_indexes],
            expected_surface_temperatures,
            atol=atol,
        )
        assert np.allclose(
            transient_result[TemperatureType.AVERAGE][useful_indexes],
            expected_mean_temperatures,
            atol=atol,
        )
        assert np.allclose(
            transient_result[TemperatureType.CORE][useful_indexes],
            expected_core_temperatures,
            atol=atol,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

from thermohl import floatArrayLike
from thermohl.power.convective_cooling import ConvectiveCoolingBase
from thermohl.power.olla import Air


class ConvectiveCooling(ConvectiveCoolingBase):
    """Convective cooling term."""

    def __init__(
        self,
        altitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        ambient_temperature: floatArrayLike,
        wind_speed: floatArrayLike,
        outer_diameter: floatArrayLike,
        wind_azimuth: floatArrayLike = None,
        wind_attack_angle: floatArrayLike = None,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Args:
            altitude (float | numpy.ndarray): Altitude (m).
            cable_azimuth (float | numpy.ndarray): Azimuth (deg).
            ambient_temperature (float | numpy.ndarray): Ambient temperature (°C).
            wind_speed (float | numpy.ndarray): Wind speed (m·s⁻¹).
            wind_azimuth (float | numpy.ndarray): wind_azimuth regarding north (deg).
            outer_diameter (float | numpy.ndarray): External diameter (m).

        """
        super().__init__(
            altitude,
            cable_azimuth,
            ambient_temperature,
            wind_speed,
            outer_diameter,
            Air.volumic_mass,
            Air.dynamic_viscosity,
            Air.thermal_conductivity,
            wind_azimuth,
            wind_attack_angle,
            **kwargs,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pytest

from thermohl.utils import bisect_v, quasi_newton_2d

_nprs = 3141592654


def test_bisect():
    np.random.seed(_nprs)
    size = 99
    tol = 1.0e-09

    c = np.linspace(1, size, size) + np.random.randn(size)
    c = np.abs(c)

    def fun(x):
        return x**2 - c

    r0 = np.sqrt(c)

    x0, err = bisect_v(
        fun,
        0.0,
        np.sqrt(np.max(c)) * 1.1,
        (size,),
        print_error=False,
        tolerance=tol,
        max_iterations=99,
    )

    assert np.max(np.abs(x0 - r0) <= np.minimum(tol, err))


def test_bisect_scalar():
    def f(x):
        return x**2 - 2

    tol = 1e-6
    x0, err = bisect_v(f, lower_bound=0, upper_bound=2, output_shape=1, tolerance=tol)
    np.testing.assert_allclose(x0, np.sqrt(2), atol=tol)


def test_bisect_vector():
    def f(x):
        return np.array([x[0] ** 2 - 2, x[1] ** 3 - 2])
        # Not the best way to use bisect_v, but just for testing.

    tol = 1e-6
    x0, err = bisect_v(
        f, lower_bound=0, upper_bound=2, output_shape=(2,), tolerance=tol
    )
    np.testing.assert_allclose(x0, np.array([np.sqrt(2), np.cbrt(2)]), atol=tol)


def test_bisect_array():
    c = np.arange(1, 28).reshape(3, 3, 3)

    def f(x):
        return x**2 - c

    tol = 1e-6
    x0, err = bisect_v(
        f, lower_bound=0, upper_bound=30, output_shape=c.shape, tolerance=tol
    )

    np.testing.assert_allclose(x0, np.sqrt(c), atol=tol)


def test_bisect_no_convergence():
    def f(x):
        return x**2 + 1  # No root

    with pytest.raises(ValueError):
        bisect_v(
            f,
            lower_bound=-50,
            upper_bound=50,
            output_shape=(1,),
        )


def test_bisect_no_convergence_array():
    c = np.arange(-1, 1)

    def f(x):
        return x**2 + c  # No root for c > 0

    with pytest.raises(ValueError):
        bisect_v(
            f,
            lower_bound=-50,
            upper_bound=50,
            output_shape=(1,),
        )


#
def test_quasi_newton_2d_convergence():
    np.random.seed(_nprs)
    size = 10
    tol = 1.0e-12

    a = np.abs(1 + np.random.randn(size))
    b = np.abs(1 + np.random.randn(size))

    def f1(x, y):
        return y - a * x**2

    def f2(x, y):
        return y - b * x**3

    xg = np.ones((size,))
    yg = np.ones((size,))
    x, y, count, err = quasi_newton_2d(
        func1=f1,
        func2=f2,
        x_init=xg,
        y_init=yg,
        relative_tolerance=tol,
        max_iterations=999,
        delta_x=1.0e-09,
        delta_y=1.0e-09,
    )

    assert np.logical_and(
        np.max(np.abs(f1(x, y))) < tol, np.max(np.abs(f2(x, y))) < tol
    )
    assert count == 999


def test_quasi_newton_2d_no_convergence():
    np.random.seed(_nprs)
    size = 10
    tol = 1.0e-12

    a = np.abs(1 + np.random.randn(size))
    b = np.abs(1 + np.random.randn(size))

    def f1(x, y):
        return y - a * x**2

    def f2(x, y):
        return y - b * x**3

    xg = np.ones((size,))
    yg = np.ones((size,))
    x, y, count, err = quasi_newton_2d(
        func1=f1,
        func2=f2,
        x_init=xg,
        y_init=yg,
        relative_tolerance=tol,
        max_iterations=1,  # Set max iterations to 1 to force no convergence
        delta_x=1.0e-09,
        delta_y=1.0e-09,
    )

    assert count == 1
    assert np.logical_or(
        np.max(np.abs(f1(x, y))) >= tol, np.max(np.abs(f2(x, y))) >= tol
    )


def test_quasi_newton_2d_large_system():
    np.random.seed(_nprs)
    size = 1000
    tol = 1.0e-12

    a = np.abs(1 + np.random.randn(size))
    b = np.abs(1 + np.random.randn(size))

    def f1(x, y):
        return y - a * x**2

    def f2(x, y):
        return y - b * x**3

    xg = np.ones((size,))
    yg = np.ones((size,))
    x, y, count, err = quasi_newton_2d(
        func1=f1,
        func2=f2,
        x_init=xg,
        y_init=yg,
        relative_tolerance=tol,
        max_iterations=999,
        delta_x=1.0e-09,
        delta_y=1.0e-09,
    )

    assert np.logical_and(
        np.max(np.abs(f1(x, y))) < tol, np.max(np.abs(f2(x, y))) < tol
    )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl.power.ieee import JouleHeating


def test_c_scalar():
    temp_low = 20.0
    temp_high = 80.0
    linear_resistance_temp_low = 0.1
    linear_resistance_temp_high = 0.2
    expected = (linear_resistance_temp_high - linear_resistance_temp_low) / (
        temp_high - temp_low
    )

    result = JouleHeating._c(
        temp_low,
        temp_high,
        linear_resistance_temp_low,
        linear_resistance_temp_high,
    )

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_c_array():
    temp_low = np.array([20.0, 30.0])
    temp_high = np.array([80.0, 90.0])
    linear_resistance_temp_low = np.array([0.1, 0.15])
    linear_resistance_temp_high = np.array([0.2, 0.25])
    expected = (linear_resistance_temp_high - linear_resistance_temp_low) / (
        temp_high - temp_low
    )

    result = JouleHeating._c(
        temp_low,
        temp_high,
        linear_resistance_temp_low,
        linear_resistance_temp_high,
    )

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_c_mixed():
    temp_low = 20.0
    temp_high = np.array([80.0, 90.0])
    linear_resistance_temp_low = 0.1
    linear_resistance_temp_high = np.array([0.2, 0.25])
    expected = (linear_resistance_temp_high - linear_resistance_temp_low) / (
        temp_high - temp_low
    )

    result = JouleHeating._c(
        temp_low,
        temp_high,
        linear_resistance_temp_low,
        linear_resistance_temp_high,
    )

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


joule_heating_instances = [
    JouleHeating(
        transit=np.array([100.0]),
        temp_low=np.array([20.0, 25.0]),
        temp_high=np.array([80.0, 85.0]),
        linear_resistance_temp_low=np.array([0.1, 0.15]),
        linear_resistance_temp_high=np.array([0.2, 0.25]),
    ),
    JouleHeating(
        transit=100.0,
        temp_low=20.0,
        temp_high=80.0,
        linear_resistance_temp_low=0.1,
        linear_resistance_temp_high=0.2,
    ),
    JouleHeating(
        transit=100.0,
        temp_low=np.array([20.0, 25.0]),
        temp_high=np.array([80.0, 85.0]),
        linear_resistance_temp_low=np.array([0.1, 0.15]),
        linear_resistance_temp_high=np.array([0.2, 0.25]),
    ),
]


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
        "JouleHeating with mixed types",
    ],
)
def test_rdc_scalar_temperature(joule_heating):
    conductor_temperature = 50.0
    expected = (
        joule_heating.linear_resistance_temp_low
        + joule_heating.temp_coeff_linear
        * (conductor_temperature - joule_heating.temp_low)
    )

    result = joule_heating._rdc(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
        "JouleHeating with mixed types",
    ],
)
def test_rdc_array_temperature(joule_heating):
    conductor_temperature = np.array([30.0, 40.0])
    expected = (
        joule_heating.linear_resistance_temp_low
        + joule_heating.temp_coeff_linear
        * (conductor_temperature - joule_heating.temp_low)
    )

    result = joule_heating._rdc(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
        "JouleHeating with mixed types",
    ],
)
def test_value_scalar_temperature(joule_heating):
    conductor_temperature = 50.0
    expected = (
        joule_heating.linear_resistance_temp_low
        + joule_heating.temp_coeff_linear
        * (conductor_temperature - joule_heating.temp_low)
    ) * joule_heating.transit**2

    result = joule_heating.value(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
        "JouleHeating with mixed types",
    ],
)
def test_value_array_temperature(joule_heating):
    conductor_temperature = np.array([30.0, 40.0])
    expected = (
        joule_heating.linear_resistance_temp_low
        + joule_heating.temp_coeff_linear
        * (conductor_temperature - joule_heating.temp_low)
    ) * joule_heating.transit**2

    result = joule_heating.value(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_value_array_temperature_different_shape_should_throw_error():
    transit = 100.0
    temp_low = np.array([20.0, 25.0])
    temp_high = np.array([80.0, 85.0])
    linear_resistance_temp_low = np.array([0.1, 0.15])
    linear_resistance_temp_high = np.array([0.2, 0.25])
    joule_heating = JouleHeating(
        transit,
        temp_low,
        temp_high,
        linear_resistance_temp_low,
        linear_resistance_temp_high,
    )
    conductor_temperature = np.array([30.0, 40.0, 50.0])

    with pytest.raises(ValueError):
        joule_heating.value(conductor_temperature)


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
        "JouleHeating with mixed types",
    ],
)
def test_derivative_scalar_temperature(joule_heating):
    conductor_temperature = 50.0
    expected = joule_heating.temp_coeff_linear * joule_heating.transit**2

    result = joule_heating.derivative(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
        "JouleHeating with mixed types",
    ],
)
def test_derivative_array_temperature(joule_heating):
    conductor_temperature = np.array([30.0, 40.0])
    expected = (
        joule_heating.temp_coeff_linear
        * joule_heating.transit**2
        * np.ones_like(conductor_temperature)
    )

    result = joule_heating.derivative(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class JouleHeating(PowerTerm):
    """Joule heating term."""

    def __init__(
        self,
        transit: floatArrayLike,
        outer_diameter: floatArrayLike,
        core_diameter: floatArrayLike,
        outer_area: floatArrayLike,
        core_area: floatArrayLike,
        magnetic_coeff: floatArrayLike,
        magnetic_coeff_per_a: floatArrayLike,
        temperature_coeff_linear: floatArrayLike,
        temperature_coeff_quadratic: floatArrayLike,
        linear_resistance_dc_20c: floatArrayLike,
        reference_temperature: floatArrayLike = 20.0,
        frequency: floatArrayLike = 50.0,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Args:
            transit (float | numpy.ndarray): Transit intensity (A).
            outer_diameter (float | numpy.ndarray): External diameter (m).
            core_diameter (float | numpy.ndarray): Core diameter (m).
            outer_area (float | numpy.ndarray): External (total) cross-sectional area (m²).
            core_area (float | numpy.ndarray): Core cross-sectional area (m²).
            magnetic_coeff (float | numpy.ndarray): Coefficient for magnetic effects (—).
            magnetic_coeff_per_a (float | numpy.ndarray): Coefficient for magnetic effects (A⁻¹).
            temperature_coeff_linear (float | numpy.ndarray): Linear resistance augmentation with temperature (K⁻¹).
            temperature_coeff_quadratic (float | numpy.ndarray): Quadratic resistance augmentation with temperature (K⁻²).
            linear_resistance_dc_20c (float | numpy.ndarray): Electric resistance per unit length (DC) at 20°C (Ω·m⁻¹).
            reference_temperature (float | numpy.ndarray, optional): Reference temperature (°C). The default is 20.
            frequency (float | numpy.ndarray, optional): Current frequency (Hz). The default is 50.

        """
        self.transit = transit
        self.outer_diameter = outer_diameter
        self.core_diameter = core_diameter
        self.magnetic_coeff = self._kem(
            outer_area, core_area, magnetic_coeff, magnetic_coeff_per_a
        )
        self.temp_coeff_linear = temperature_coeff_linear
        self.temp_coeff_quadratic = temperature_coeff_quadratic
        self.linear_resistance_dc_20c = linear_resistance_dc_20c
        self.reference_temperature = reference_temperature
        self.frequency = frequency

    def _rdc(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        """
        Compute resistance per unit length for direct current.

        Args:
            conductor_temperature (float | numpy.ndarray): Temperature at which to compute the resistance (°C).

        Returns:
            float | numpy.ndarray: Resistance per unit length for direct current at the given temperature(s) (Ω·m⁻¹).
        """
        temperature_delta = conductor_temperature - self.reference_temperature
        return self.linear_resistance_dc_20c * (
            1.0
            + self.temp_coeff_linear * temperature_delta
            + self.temp_coeff_quadratic * temperature_delta**2
        )

    def _ks(self, dc_resistance: floatArrayLike) -> floatArrayLike:
        """
        Compute skin-effect coefficient.

        This method calculates the skin-effect coefficient based on the given
        resistance (rdc) and the object's attributes. The calculation is an
        approximation as described in the RTE's document.

        Args:
            dc_resistance (float | numpy.ndarray): The resistance value(s) for which the skin-effect coefficient is to be computed (Ω·m⁻¹).

        Returns:
            floatArrayLike: The computed skin-effect coefficient(s) (—).
        """
        skin_param = (
            8
            * np.pi
            * self.frequency
            * (self.outer_diameter - self.core_diameter) ** 2
            / (
                (self.outer_diameter**2 - self.core_diameter**2)
                * 1.0e07
                * dc_resistance
            )
        )
        coeff_a = 7 * skin_param**2 / (315 + 3 * skin_param**2)
        coeff_b = 56 / (211 + skin_param**2)
        core_ratio = 1.0 - self.core_diameter / self.outer_diameter
        return 1.0 + coeff_a * (1.0 - 0.5 * core_ratio - coeff_b * core_ratio**2)

    def _kem(
        self,
        outer_area: floatArrayLike,
        core_area: floatArrayLike,
        magnetic_coeff: floatArrayLike,
        magnetic_coeff_per_a: floatArrayLike,
    ) -> floatArrayLike:
        """
        Compute magnetic coefficient.

        Args:
            outer_area (float | numpy.ndarray): External (total) cross-sectional area (m²).
            core_area (float | numpy.ndarray): Core cross-sectional area (m²).
            magnetic_coeff (float | numpy.ndarray): Coefficient for magnetic effects (—).
            magnetic_coeff_per_a (float | numpy.ndarray): Coefficient for magnetic effects (A⁻¹).

        Returns:
            floatArrayLike: Computed magnetic coefficient (—).
        """
        scale = (
            np.ones_like(self.transit)
            * np.ones_like(outer_area)
            * np.ones_like(core_area)
            * np.ones_like(magnetic_coeff)
            * np.ones_like(magnetic_coeff_per_a)
        )
        is_scalar = scale.shape == ()
        if is_scalar:
            scale = np.array([1.0])
        current = self.transit * scale
        core_area = core_area * scale
        outer_area = outer_area * scale
        has_core = core_area > 0.0
        magnetic_slope = magnetic_coeff_per_a * scale
        magnetic_coeff = magnetic_coeff * scale
        magnetic_coeff[has_core] += (
            magnetic_slope[has_core]
            * current[has_core]
            / ((outer_area[has_core] - core_area[has_core]) * 1.0e06)
        )
        if is_scalar:
            magnetic_coeff = magnetic_coeff[0]
        return magnetic_coeff

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        dc_resistance = self._rdc(conductor_temperature)
        skin_effect_coeff = self._ks(dc_resistance)
        ac_resistance = self.magnetic_coeff * skin_effect_coeff * dc_resistance
        return ac_resistance * self.transit**2



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl.power.cigre import Air


def test_volumic_mass_scalar():
    air_temperature = 25.0
    altitude = 1000.0
    expected = 1.2925 * np.exp(-1.16e-04 * altitude)

    result = Air.volumic_mass(air_temperature, altitude)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_volumic_mass_array():
    air_temperature = np.array([15.0, 20.0, 25.0])
    altitude = np.array([0.0, 500.0, 1000.0])
    expected = 1.2925 * np.exp(-1.16e-04 * altitude)

    result = Air.volumic_mass(air_temperature, altitude)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_volumic_mass_default_altitude():
    air_temperature = np.array([15.0, 20.0, 25.0])
    expected = 1.2925 * np.ones_like(air_temperature)

    result = Air.volumic_mass(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_volumic_mass_mismatched_array_sizes():
    air_temperature = np.array([15.0, 20.0])
    altitude = np.array([0.0, 500.0, 1000.0])
    with pytest.raises(ValueError):
        Air.volumic_mass(air_temperature, altitude)


def test_relative_density_scalar():
    air_temperature = 25.0
    altitude = 1000.0
    expected = np.exp(-1.16e-04 * altitude)

    result = Air.relative_density(air_temperature, altitude)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_relative_density_array():
    air_temperature = np.array([15.0, 20.0, 25.0])
    altitude = np.array([0.0, 500.0, 1000.0])
    expected = np.exp(-1.16e-04 * altitude)

    result = Air.relative_density(air_temperature, altitude)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_relative_density_default_altitude():
    air_temperature = np.array([15.0, 20.0, 25.0])
    expected = np.ones_like(air_temperature)

    result = Air.relative_density(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_relative_density_mismatched_array_sizes():
    air_temperature = np.array([15.0, 20.0])
    altitude = np.array([0.0, 500.0, 1000.0])
    with pytest.raises(ValueError):
        Air.relative_density(air_temperature, altitude)


def test_kinematic_viscosity_scalar():
    air_temperature = 25.0
    expected = 1.32e-05 + 9.5e-08 * air_temperature

    result = Air.kinematic_viscosity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_kinematic_viscosity_array():
    air_temperature = np.array([15.0, 20.0, 25.0])
    expected = 1.32e-05 + 9.5e-08 * air_temperature

    result = Air.kinematic_viscosity(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_kinematic_viscosity_negative_temperature():
    air_temperature = -10.0
    expected = 1.32e-05 + 9.5e-08 * air_temperature

    result = Air.kinematic_viscosity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_kinematic_viscosity_zero_temperature():
    air_temperature = 0.0
    expected = 1.32e-05

    result = Air.kinematic_viscosity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_dynamic_viscosity_scalar():
    air_temperature = 25.0
    altitude = 1000.0
    expected = Air.kinematic_viscosity(air_temperature) * Air.volumic_mass(
        air_temperature, altitude
    )

    result = Air.dynamic_viscosity(air_temperature, altitude)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_dynamic_viscosity_array():
    air_temperature = np.array([15.0, 20.0, 25.0])
    altitude = np.array([0.0, 500.0, 1000.0])
    expected = Air.kinematic_viscosity(air_temperature) * Air.volumic_mass(
        air_temperature, altitude
    )

    result = Air.dynamic_viscosity(air_temperature, altitude)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_dynamic_viscosity_default_altitude():
    air_temperature = np.array([15.0, 20.0, 25.0])
    expected = Air.kinematic_viscosity(air_temperature) * Air.volumic_mass(
        air_temperature
    )

    result = Air.dynamic_viscosity(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_dynamic_viscosity_mismatched_array_sizes():
    air_temperature = np.array([15.0, 20.0])
    altitude = np.array([0.0, 500.0, 1000.0])
    with pytest.raises(ValueError):
        Air.dynamic_viscosity(air_temperature, altitude)


def test_thermal_conductivity_scalar():
    air_temperature = 25.0
    expected = 2.42e-02 + 7.2e-05 * air_temperature

    result = Air.thermal_conductivity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_thermal_conductivity_array():
    air_temperature = np.array([15.0, 20.0, 25.0])
    expected = 2.42e-02 + 7.2e-05 * air_temperature

    result = Air.thermal_conductivity(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_thermal_conductivity_negative_temperature():
    air_temperature = -10.0
    expected = 2.42e-02 + 7.2e-05 * air_temperature

    result = Air.thermal_conductivity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_thermal_conductivity_zero_temperature():
    air_temperature = 0.0
    expected = 2.42e-02

    result = Air.thermal_conductivity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_prandtl_scalar():
    air_temperature = 25.0
    expected = 0.715 - 2.5e-04 * air_temperature

    result = Air.prandtl(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_prandtl_array():
    air_temperature = np.array([15.0, 20.0, 25.0])
    expected = 0.715 - 2.5e-04 * air_temperature

    result = Air.prandtl(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_prandtl_negative_temperature():
    air_temperature = -10.0
    expected = 0.715 - 2.5e-04 * air_temperature

    result = Air.prandtl(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_prandtl_zero_temperature():
    air_temperature = 0.0
    expected = 0.715

    result = Air.prandtl(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
import numpy as np

from thermohl.power.ieee import SolarHeating


def test_solar_heating_init_scalar():
    latitude = 45.0
    altitude = 1000.0
    cable_azimuth = 180.0
    turbidity = 0.5
    datetime_utc = datetime(2000, 6, 21, 12, tzinfo=timezone.utc)
    outer_diameter = 0.01
    solar_absorptivity = 0.9
    srad = 800.0

    solar_heating = SolarHeating(
        latitude,
        altitude,
        cable_azimuth,
        turbidity,
        datetime_utc,
        outer_diameter,
        solar_absorptivity,
        srad,
    )

    assert solar_heating.solar_absorptivity == solar_absorptivity
    assert np.isclose(solar_heating.solar_irradiance, srad)
    assert np.isclose(solar_heating.outer_diameter, outer_diameter)


def test_solar_heating_init_array():
    latitude = np.array([45.0, 50.0])
    altitude = np.array([1000.0, 2000.0])
    cable_azimuth = np.array([180.0, 190.0])
    turbidity = np.array([0.5, 0.7])
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 7, 22, 13, tzinfo=timezone.utc),
    ]
    outer_diameter = np.array([0.01, 0.02])
    solar_absorptivity = np.array([0.9, 0.8])
    srad = np.array([800.0, 900.0])

    solar_heating = SolarHeating(
        latitude,
        altitude,
        cable_azimuth,
        turbidity,
        datetime_utc,
        outer_diameter,
        solar_absorptivity,
        srad,
    )

    assert np.allclose(solar_heating.solar_absorptivity, solar_absorptivity)
    assert np.allclose(solar_heating.solar_irradiance, srad)
    assert np.allclose(solar_heating.outer_diameter, outer_diameter)


def test_solar_heating_init_mixed():
    latitude = 45.0
    altitude = 1000.0
    cable_azimuth = 180.0
    turbidity = 0.5
    datetime_utc = datetime(2000, 6, 21, 12, tzinfo=timezone.utc)
    outer_diameter = 0.01
    solar_absorptivity = 0.9
    srad = np.array([800.0, 900.0])

    solar_heating = SolarHeating(
        latitude,
        altitude,
        cable_azimuth,
        turbidity,
        datetime_utc,
        outer_diameter,
        solar_absorptivity,
        srad,
    )

    assert solar_heating.solar_absorptivity == solar_absorptivity
    assert np.allclose(solar_heating.solar_irradiance, srad)
    assert np.isclose(solar_heating.outer_diameter, outer_diameter)


def test_solar_heating_init_no_srad():
    latitude = 45.0
    altitude = 1000.0
    cable_azimuth = 180.0
    turbidity = 0.5
    datetime_utc = datetime(2000, 6, 21, 12, tzinfo=timezone.utc)
    outer_diameter = 0.01
    solar_absorptivity = 0.9

    solar_heating = SolarHeating(
        latitude,
        altitude,
        cable_azimuth,
        turbidity,
        datetime_utc,
        outer_diameter,
        solar_absorptivity,
        np.nan,
    )

    assert solar_heating.solar_absorptivity == solar_absorptivity
    assert solar_heating.solar_irradiance is not None
    assert np.isclose(solar_heating.outer_diameter, outer_diameter)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Base class to build a solver for heat equation."""

from datetime import timedelta
from abc import ABC, abstractmethod
from typing import Type, Any, Optional, Dict, Iterable
import numpy.typing as npt
import numpy as np
import pandas as pd
from numpy import ndarray

from thermohl import (
    floatArrayLike,
    floatArray,
    numberArray,
    numberArrayLike,
    datetimeListLike,
)
from thermohl.power import PowerTerm
from thermohl.solver.parameters import Parameters
from thermohl.solver.entities import PowerType


class Solver(ABC):
    """Object to solve a temperature problem.

    The temperature of a conductor is driven by four power terms, two heating
    terms (joule and solar heating) and three cooling terms (convective,
    radiative and precipitation cooling). This class is used to solve a
    temperature problem with the heating and cooling terms passed to its
    __init__ function.
    """

    def __init__(
        self,
        dic: Optional[dict[str, Any]] = None,
        joule: Type[PowerTerm] = PowerTerm,
        solar: Type[PowerTerm] = PowerTerm,
        convective: Type[PowerTerm] = PowerTerm,
        radiative: Type[PowerTerm] = PowerTerm,
        precipitation: Type[PowerTerm] = PowerTerm,
    ) -> None:
        """Create a Solver object.

        Args:
            dic (dict[str, Any] | None): Input values used in power terms. If there is a missing value, a default is used.
            joule (Type[PowerTerm]): Joule heating term class.
            solar (Type[PowerTerm]): Solar heating term class.
            convective (Type[PowerTerm]): Convective cooling term class.
            radiative (Type[PowerTerm]): Radiative cooling term class.
            precipitation (Type[PowerTerm]): Precipitation cooling term class.

        Returns:
            None

        """
        self.args = Parameters(dic)
        self.args.extend()
        self.joule_heating = joule(**self.args.__dict__)
        self.solar_heating = solar(**self.args.__dict__)
        self.convective_cooling = convective(**self.args.__dict__)
        self.radiative_cooling = radiative(**self.args.__dict__)
        self.precipitation_cooling = precipitation(**self.args.__dict__)
        self.args.compress()

    def update(self) -> None:
        self.args.extend()
        self.joule_heating.__init__(**self.args.__dict__)
        self.solar_heating.__init__(**self.args.__dict__)
        self.convective_cooling.__init__(**self.args.__dict__)
        self.radiative_cooling.__init__(**self.args.__dict__)
        self.precipitation_cooling.__init__(**self.args.__dict__)
        self.args.compress()

    def balance(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        return (
            self.joule_heating.value(conductor_temperature)
            + self.solar_heating.value(conductor_temperature)
            - self.convective_cooling.value(conductor_temperature)
            - self.radiative_cooling.value(conductor_temperature)
            - self.precipitation_cooling.value(conductor_temperature)
        )

    @staticmethod
    def powers() -> tuple[PowerType, PowerType, PowerType, PowerType, PowerType]:
        return (
            PowerType.JOULE,
            PowerType.SOLAR,
            PowerType.CONVECTION,
            PowerType.RADIATION,
            PowerType.RAIN,
        )

    @abstractmethod
    def steady_temperature(self) -> pd.DataFrame:
        raise NotImplementedError

    @abstractmethod
    def transient_temperature(self) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def steady_intensity(self) -> pd.DataFrame:
        raise NotImplementedError


def reshape(input_array: numberArrayLike, nb_row: int, nb_columns: int) -> numberArray:
    """
    Reshape the input array to the specified dimensions (nr, nc) if possible.

    Args:
        input_array (numberArrayLike): Input array to be reshaped.
        nb_row (int): Desired number of rows for the reshaped array.
        nb_columns (int): Desired number of columns for the reshaped array.

    Returns:
        numberArray: Reshaped array of size (nb_row, nb_columns). If reshaping is not possible,
            returns an array filled with the input_value repeated to fill the dimension (nb_row, nb_columns).

    Raises:
        AttributeError: If the input_array has an invalid shape that cannot be reshaped.
    """
    reshaped_array = ndarray
    try:
        input_shape = input_array.shape
        if len(input_shape) == 1:
            if nb_row == input_shape[0]:
                reshaped_array = np.column_stack(nb_columns * (input_array,))
            elif nb_columns == input_shape[0]:
                reshaped_array = np.vstack(nb_row * (input_array,))
        elif len(input_shape) == 0:
            raise AttributeError()
        else:
            reshaped_array = np.reshape(input_array, (nb_row, nb_columns))
    except AttributeError:
        reshaped_array = input_array * np.ones(
            (nb_row, nb_columns), dtype=type(input_array)
        )
    return reshaped_array


def _set_dates(
    datetime_utc: datetimeListLike,
    offset: floatArray,
    number_of_computations: int,
) -> npt.NDArray[np.datetime64]:
    """
    This function is used in transient temperature computations
    It provides a 2D array of size (len(offset), number_of_computations) such that
    datetime_with_offset[i, j] = datetime_utc[j] + offset[i]

    :param datetime_utc: Datetime or list of datetimes representing the initial times.
    :param offset: Array representing the time offsets in seconds.
    :param number_of_computations: Number of computations.

    :return: 2D array of shape (len(offset), number_of_computations) with datetime values.
    """
    datetime_utc = (
        np.array(datetime_utc)
        if isinstance(datetime_utc, Iterable)
        else np.array(number_of_computations * [datetime_utc])
    )

    number_of_offset = len(offset)
    datetime_with_offset = np.zeros(
        (number_of_offset, number_of_computations), dtype=object
    )
    for i in range(number_of_offset):
        datetime_with_offset[i, :] = datetime_utc + timedelta(seconds=float(offset[i]))

    return datetime_with_offset


def get_time_changing_parameters(args, offset, N, n):
    # get datetime for each offset
    datetime_utc = _set_dates(args.datetime_utc, offset, n)

    # A dict with time-changing quantities (with all elements of size N * n)
    de = {
        "datetime_utc": datetime_utc,
        "transit": reshape(args.transit, N, n),
        "ambient_temperature": reshape(args.ambient_temperature, N, n),
        "wind_azimuth": reshape(args.wind_azimuth, N, n),
        "wind_speed": reshape(args.wind_speed, N, n),
        "ambient_pressure": reshape(args.ambient_pressure, N, n),
        "relative_humidity": reshape(args.relative_humidity, N, n),
        "precipitation_rate": reshape(args.precipitation_rate, N, n),
    }
    del datetime_utc
    return de



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Callable, Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


def compute_wind_attack_angle(
    cable_azimuth: floatArrayLike, wind_azimuth: floatArrayLike
) -> floatArrayLike:
    """
    Compute wind attack angle.

    Args:
        cable_azimuth (float | numpy.ndarray): Cable azimuth (deg).
        wind_azimuth (float | numpy.ndarray): Wind azimuth regarding north (deg).

    Returns:
        float | numpy.ndarray: Wind attack angle (rad).
    """
    return np.arcsin(np.sin(np.deg2rad(np.abs(cable_azimuth - wind_azimuth) % 180.0)))


class ConvectiveCoolingBase(PowerTerm):
    """Convective cooling term."""

    def __init__(
        self,
        altitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        ambient_temperature: floatArrayLike,
        wind_speed: floatArrayLike,
        outer_diameter: floatArrayLike,
        air_density: Callable[[floatArrayLike, floatArrayLike], floatArrayLike],
        dynamic_viscosity: Callable[[floatArrayLike], floatArrayLike],
        thermal_conductivity: Callable[[floatArrayLike], floatArrayLike],
        wind_azimuth: floatArrayLike = None,
        wind_attack_angle: floatArrayLike = None,
        **kwargs: Any,
    ):
        self.altitude = altitude
        self.ambient_temp = ambient_temperature
        self.wind_speed = wind_speed

        if wind_attack_angle is None and wind_azimuth is None:
            raise ValueError("Must provide either wind_attack_angle or wind_azimuth.")
        if wind_attack_angle is not None and wind_azimuth is not None:
            print(
                "Warning: both wind_attack_angle and wind_azimuth are provided. wind_azimuth will be ignored."
            )
        if wind_attack_angle is not None:
            self.wind_attack_angle = wind_attack_angle
        else:
            self.wind_attack_angle = compute_wind_attack_angle(
                cable_azimuth, wind_azimuth
            )

        self.outer_diameter = outer_diameter

        self.air_density = air_density
        self.dynamic_viscosity = dynamic_viscosity
        self.thermal_conductivity = thermal_conductivity

    def _value_forced(
        self,
        film_temperature: floatArrayLike,
        temperature_delta: floatArrayLike,
        air_density: floatArrayLike,
    ) -> floatArrayLike:
        """
        Compute forced convective cooling value.

        Args:
            film_temperature (float | numpy.ndarray): Film temperature (°C).
            temperature_delta (float | numpy.ndarray): Temperature difference (°C).
            air_density (float | numpy.ndarray): Velocity magnitude proxy (relative density or similar, model-dependent).

        Returns:
            float | numpy.ndarray: Computed forced convective cooling values (W·m⁻¹).
        """
        reynolds = (
            self.wind_speed
            * self.outer_diameter
            * air_density
            / self.dynamic_viscosity(film_temperature)
        )
        direction_factor = (
            1.194
            - np.cos(self.wind_attack_angle)
            + 0.194 * np.cos(2.0 * self.wind_attack_angle)
            + 0.368 * np.sin(2.0 * self.wind_attack_angle)
        )
        return (
            direction_factor
            * np.maximum(1.01 + 1.35 * reynolds**0.52, 0.754 * reynolds**0.6)
            * self.thermal_conductivity(film_temperature)
            * temperature_delta
        )

    def _value_natural(
        self,
        temperature_delta: floatArrayLike,
        air_density: floatArrayLike,
    ) -> floatArrayLike:
        """
        Compute natural convective cooling value.

        Args:
            temperature_delta (float | numpy.ndarray): Temperature difference (°C).
            air_density (float | numpy.ndarray): Velocity magnitude (relative density or similar, model-dependent).

        Returns:
            float | numpy.ndarray: Natural convective cooling value (W·m⁻¹).
        """
        return (
            3.645
            * np.sqrt(air_density)
            * self.outer_diameter**0.75
            * np.sign(temperature_delta)
            * np.abs(temperature_delta) ** 1.25
        )

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute convective cooling.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        film_temperature = 0.5 * (conductor_temperature + self.ambient_temp)
        temperature_delta = conductor_temperature - self.ambient_temp
        air_density = self.air_density(film_temperature, self.altitude)
        return np.maximum(
            self._value_forced(film_temperature, temperature_delta, air_density),
            self._value_natural(temperature_delta, air_density),
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import List, Optional, Any, Iterable

import numpy as np

from thermohl import floatArrayLike, sun, datetimeListLike
from thermohl.power.power_term import PowerTerm
from thermohl.sun import time_to_float_hours


class _SRad:
    """Solar radiation calculator."""

    def __init__(self, clean: List[float], indus: List[float]):
        """Initialize the solar radiation calculator.

        :param clean: Coefficients for the polynomial function to compute atmospheric turbidity in clean air conditions.
        :param indus: Coefficients for the polynomial function to compute atmospheric turbidity in industrial (polluted) air conditions.
        """
        self.clean = clean
        self.indus = indus

    def atmosphere_turbidity(
        self,
        solar_altitude: floatArrayLike,
        turbidity: Optional[floatArrayLike] = 0.0,
    ) -> floatArrayLike:
        """Compute coefficient for atmosphere turbidity.
        This method calculates the atmospheric turbidity coefficient using a polynomial
        function of the solar altitude. The coefficients of the polynomial are a weighted
        average of the clean air and industrial air coefficients, with the weights
        determined by the turbidity factor.

        :param solar_altitude: Solar altitude in degrees.
        :param turbidity: Atmospheric turbidity factor (0 for clean air, 1 for industrial air).
        :return: Coefficient for atmospheric turbidity.
        """
        clean_weight = 1.0 - turbidity
        coeff_6 = clean_weight * self.clean[6] + turbidity * self.indus[6]
        coeff_5 = clean_weight * self.clean[5] + turbidity * self.indus[5]
        coeff_4 = clean_weight * self.clean[4] + turbidity * self.indus[4]
        coeff_3 = clean_weight * self.clean[3] + turbidity * self.indus[3]
        coeff_2 = clean_weight * self.clean[2] + turbidity * self.indus[2]
        coeff_1 = clean_weight * self.clean[1] + turbidity * self.indus[1]
        coeff_0 = clean_weight * self.clean[0] + turbidity * self.indus[0]
        return (
            coeff_6 * solar_altitude**6
            + coeff_5 * solar_altitude**5
            + coeff_4 * solar_altitude**4
            + coeff_3 * solar_altitude**3
            + coeff_2 * solar_altitude**2
            + coeff_1 * solar_altitude**1
            + coeff_0
        )

    def __call__(
        self,
        latitude: floatArrayLike,
        altitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        turbidity: floatArrayLike,
        datetime_utc: datetimeListLike,
    ) -> floatArrayLike:
        """Compute solar radiation."""
        date = (
            [d.date() for d in datetime_utc]
            if isinstance(datetime_utc, Iterable)
            else datetime_utc.date()
        )
        hour = (
            np.array([time_to_float_hours(d.time()) for d in datetime_utc])
            if isinstance(datetime_utc, Iterable)
            else time_to_float_hours(datetime_utc.time())
        )
        computed_solar_altitude = sun.solar_altitude(latitude, date, hour)
        computed_solar_azimuth = sun.solar_azimuth(latitude, date, hour)
        computed_incidence_angle = np.arccos(
            np.cos(computed_solar_altitude)
            * np.cos(computed_solar_azimuth - cable_azimuth)
        )
        altitude_factor = 1.0 + 1.148e-04 * altitude - 1.108e-08 * altitude**2
        clearness_factor = self.atmosphere_turbidity(
            np.rad2deg(computed_solar_altitude), turbidity
        )
        solar_irradiance = (
            altitude_factor * clearness_factor * np.sin(computed_incidence_angle)
        )
        return np.where(solar_irradiance > 0.0, solar_irradiance, 0.0)


class SolarHeatingBase(PowerTerm):
    """Solar heating term."""

    def __init__(
        self,
        latitude: floatArrayLike,
        altitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        turbidity: floatArrayLike,
        datetime_utc: datetimeListLike,
        outer_diameter: floatArrayLike,
        solar_absorptivity: floatArrayLike,
        est: _SRad,
        measured_solar_irradiance: floatArrayLike,
        **kwargs: Any,
    ):
        self.solar_absorptivity = solar_absorptivity

        mask = np.isnan(measured_solar_irradiance)
        self.solar_irradiance = np.empty_like(measured_solar_irradiance)
        if np.any(~mask):
            self.solar_irradiance[~mask] = np.maximum(measured_solar_irradiance, 0.0)
        if np.any(mask):
            self.solar_irradiance[mask] = est(
                np.deg2rad(latitude),
                altitude,
                np.deg2rad(cable_azimuth),
                turbidity,
                datetime_utc,
            )

        self.outer_diameter = outer_diameter

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        """Compute solar heating.

        :param conductor_temperature: Conductor temperature (°C).
        :return: Power term value (W·m⁻¹).
        """
        return (
            self.solar_absorptivity
            * self.solar_irradiance
            * self.outer_diameter
            * np.ones_like(conductor_temperature)
        )

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        """Compute solar heating derivative.

        :param conductor_temperature: Conductor temperature.
        :return: Derivative of solar heating.
        """
        return np.zeros_like(conductor_temperature)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl.power.rte import JouleHeating

joule_heating_instances = [
    JouleHeating(
        transit=np.array([10.0]),
        outer_diameter=np.array([0.01]),
        core_diameter=np.array([0.005]),
        outer_area=np.array([0.0001]),
        core_area=np.array([0.00005]),
        magnetic_coeff=np.array([1.0]),
        magnetic_coeff_per_a=np.array([0.1]),
        temperature_coeff_linear=np.array([0.004]),
        temperature_coeff_quadratic=np.array([0.0001]),
        linear_resistance_dc_20c=np.array([0.02]),
        reference_temperature=20.0,
        frequency=50.0,
    ),
    JouleHeating(
        transit=10.0,
        outer_diameter=0.01,
        core_diameter=0.005,
        outer_area=0.0001,
        core_area=0.00005,
        magnetic_coeff=1.0,
        magnetic_coeff_per_a=0.1,
        temperature_coeff_linear=0.004,
        temperature_coeff_quadratic=0.0001,
        linear_resistance_dc_20c=0.02,
        reference_temperature=20.0,
        frequency=50.0,
    ),
]


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=["JouleHeating with arrays", "JouleHeating with scalars"],
)
def test_rdc(joule_heating):
    conductor_temperature = np.array([30.0])
    expected_rdc = 0.021

    result = joule_heating._rdc(conductor_temperature)

    np.testing.assert_allclose(result, expected_rdc, rtol=1e-5)


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=["JouleHeating with arrays", "JouleHeating with scalars"],
)
def test_ks(joule_heating):
    conductor_temperature = np.array([30.0])

    rdc = joule_heating._rdc(conductor_temperature)
    expected_ks = 1.0

    result = joule_heating._ks(rdc)

    np.testing.assert_allclose(result, expected_ks, rtol=1e-5)


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=["JouleHeating with arrays", "JouleHeating with scalars"],
)
def test_kem(joule_heating):
    outer_area = np.array([0.0001])
    core_area = np.array([0.00005])
    magnetic_coeff = np.array([1.0])
    magnetic_coeff_per_a = np.array([0.1])
    expected_kem = 1.02

    result = joule_heating._kem(
        outer_area, core_area, magnetic_coeff, magnetic_coeff_per_a
    )

    np.testing.assert_allclose(result, expected_kem, rtol=1e-5)


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=["JouleHeating with arrays", "JouleHeating with scalars"],
)
def test_joule_heating_value(joule_heating):
    conductor_temperature = np.array([30.0])
    expected_value = 2.1420

    result = joule_heating.value(conductor_temperature)

    np.testing.assert_allclose(result, expected_value, rtol=1e-5)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Tuple, Type, Optional, Dict, Any, Callable

import numpy as np
import pandas as pd

from thermohl import floatArrayLike, floatArray, intArray
from thermohl.power import PowerTerm
from thermohl.solver.solver import (
    Solver as Solver_,
    get_time_changing_parameters,
)
from thermohl.solver.parameters import DEFAULT_PARAMETERS as default
from thermohl.solver.entities import (
    TargetType,
    CableLocationListLike,
    CableType,
    CableTypeListLike,
    PowerType,
    TemperatureType,
    VariableType,
)
from thermohl.solver.slv1t import Solver1T
from thermohl.utils import quasi_newton_2d


def _profile_mom(
    surface_temperature: float,
    core_temperature: float,
    radius: floatArrayLike,
    outer_radius: float,
) -> floatArrayLike:
    """Analytic temperature profile for steady heat equation in cylinder (mono-mat)."""
    return surface_temperature + (core_temperature - surface_temperature) * (
        1.0 - (radius / outer_radius) ** 2
    )


def _phi(
    radius: floatArrayLike, core_radius: floatArrayLike, outer_radius: floatArrayLike
) -> floatArrayLike:
    """Primitive function used in _profile_bim*** functions."""
    core_radius_2 = core_radius**2
    return (
        0.5 * (radius**2 - core_radius_2) - core_radius_2 * np.log(radius / core_radius)
    ) / (outer_radius**2 - core_radius_2)


def _profile_bim_avg_coeffs(
    core_radius: floatArrayLike, outer_radius: floatArrayLike
) -> tuple[floatArrayLike, floatArrayLike]:
    core_radius_2 = core_radius**2
    outer_radius_2 = outer_radius**2
    a = (
        0.5 * (outer_radius_2 - core_radius_2) ** 2
        - outer_radius_2
        * core_radius_2
        * (2.0 * np.log(outer_radius / core_radius) - 1.0)
        - core_radius**4
    )
    b = (
        2.0
        * outer_radius_2
        * (outer_radius_2 - core_radius_2)
        * _phi(outer_radius, core_radius, outer_radius)
    )
    return a, b


def _profile_bim_avg(
    surface_temperature: floatArrayLike,
    core_temperature: floatArrayLike,
    core_radius: floatArrayLike,
    outer_radius: floatArrayLike,
) -> floatArrayLike:
    """Analytical formulation for average temperature in _profile_bim."""
    a, b = _profile_bim_avg_coeffs(core_radius, outer_radius)
    return core_temperature - (a / b) * (core_temperature - surface_temperature)


def _infer_target_from_cable_type(
    cable_type: CableTypeListLike,
    target: CableLocationListLike,
) -> CableLocationListLike:
    """Infer target cable location from cable type: HOMOGENEOUS -> AVERAGE, BIMETALLIC -> CORE.

    If both target and cable_type are provided, target is ignored."""

    if target is not None and cable_type is not None:
        print(
            "WARNING: Both target and cable_type are provided. Ignoring given target and using cable_type to determine target instead."
        )

    if cable_type is None:
        return target

    if isinstance(cable_type, CableType):
        if cable_type == CableType.HOMOGENEOUS:
            return TargetType.AVERAGE
        elif cable_type == CableType.BIMETALLIC:
            return TargetType.CORE
    else:
        return [
            TargetType.AVERAGE if ct == CableType.HOMOGENEOUS else TargetType.CORE
            for ct in cable_type
        ]


class Solver3T(Solver_):
    def __init__(
        self,
        dic: Optional[dict[str, Any]] = None,
        joule: Type[PowerTerm] = PowerTerm,
        solar: Type[PowerTerm] = PowerTerm,
        convective: Type[PowerTerm] = PowerTerm,
        radiative: Type[PowerTerm] = PowerTerm,
        precipitation: Type[PowerTerm] = PowerTerm,
    ):
        super().__init__(dic, joule, solar, convective, radiative, precipitation)
        self.update()

    def _morgan_coefficients(
        self,
    ) -> Tuple[floatArray, floatArray, floatArray, intArray]:
        """
        Calculate coefficients for heat flux between surface and core in steady state.

        Returns:
        --------
        Tuple[numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[int]]
            - heat_capacity : numpy.ndarray[float]
                Coefficient array for heat flux.
            - D_ : numpy.ndarray[float]
                Array of core diameters, broadcasted to the shape of `heat_capacity`.
            - d_ : numpy.ndarray[float]
                Array of surface diameters, broadcasted to the shape of `heat_capacity`.
            - positive_surface_diameter_indices : numpy.ndarray[int]
                Indices where surface diameter `d_` is greater than 0.
        """
        heat_capacity = 0.5 * np.ones((self.args.get_number_of_computations(),))
        outer_diameter = self.args.outer_diameter * np.ones_like(heat_capacity)
        core_diameter = self.args.core_diameter * np.ones_like(heat_capacity)
        positive_surface_diameter_indices = np.nonzero(core_diameter > 0.0)[0]
        heat_capacity[positive_surface_diameter_indices] -= (
            core_diameter[positive_surface_diameter_indices] ** 2
            / (
                outer_diameter[positive_surface_diameter_indices] ** 2
                - core_diameter[positive_surface_diameter_indices] ** 2
            )
        ) * np.log(
            outer_diameter[positive_surface_diameter_indices]
            / core_diameter[positive_surface_diameter_indices]
        )
        return (
            heat_capacity,
            outer_diameter,
            core_diameter,
            positive_surface_diameter_indices,
        )

    def update(self) -> None:
        """
        Updates the solver's internal state by reinitializing several components
        and recalculating the Morgan coefficients.
        This method performs the following steps:
        1. Extends the arguments to their maximum length.
        2. Reinitializes the `joule_heating`, `solar_heating`, `convective_cooling`, `radiative_cooling`, and `precipitation_cooling` components using the updated arguments.
        3. Recalculates the Morgan coefficients using the updated dimensions.
        4. Compresses the arguments.
        Returns:
            None
        """
        self.args.extend()
        self.joule_heating.__init__(**self.args.__dict__)
        self.solar_heating.__init__(**self.args.__dict__)
        self.convective_cooling.__init__(**self.args.__dict__)
        self.radiative_cooling.__init__(**self.args.__dict__)
        self.precipitation_cooling.__init__(**self.args.__dict__)
        self.morgan_coefficients = self._morgan_coefficients()
        self.args.compress()

    def average(
        self, surface_temperature: floatArray, core_temperature: floatArray
    ) -> floatArrayLike:
        """
        Compute average temperature given surface and core temperature.

        This formula is based on analytical solution in steady-state mode. For
        single material, the formula reduces itself to an usual mean; for
        bi-material conductors, we have geometrical terms to take into account.

        Args:
            surface_temperature (numpy.ndarray): Array of surface temperatures.
            core_temperature (numpy.ndarray): Array of core temperatures.

        Returns:
            float | numpy.ndarray: Array of average temperatures.
        """
        average_temperature = 0.5 * (surface_temperature + core_temperature)
        _, outer_diameter, core_diameter, positive_surface_diameter_indices = (
            self.morgan_coefficients
        )
        average_temperature[positive_surface_diameter_indices] = _profile_bim_avg(
            surface_temperature[positive_surface_diameter_indices],
            core_temperature[positive_surface_diameter_indices],
            0.5 * core_diameter[positive_surface_diameter_indices],
            0.5 * outer_diameter[positive_surface_diameter_indices],
        )
        return average_temperature

    def joule(
        self, surface_temperature: floatArray, core_temperature: floatArray
    ) -> floatArrayLike:
        """
        Calculate the Joule heating effect.

        Args:
            surface_temperature (numpy.ndarray): Array of surface temperatures.
            core_temperature (numpy.ndarray): Array of core temperatures.

        Returns:
            float | numpy.ndarray: The calculated Joule heating values.

        Notes:
        - The function computes the average temperature `temperature`.
        - Returns the Joule heating values based on the adjusted temperatures.
        """
        average_temperature = self.average(surface_temperature, core_temperature)
        return self.joule_heating.value(average_temperature)

    def balance(
        self, surface_temperature: floatArray, core_temperature: floatArray
    ) -> floatArrayLike:
        """
        Calculate the thermal balance.

        This method computes the thermal balance by summing the joule heating,
        specific heat, and subtracting the contributions from the cooling
        components (convection, radiation, and conduction).

        Args:
            surface_temperature (numpy.ndarray): Array of surface temperatures.
            core_temperature (numpy.ndarray): Array of core temperatures.

        Returns:
            float | numpy.ndarray: The resulting thermal balance.
        """
        return (
            self.joule(surface_temperature, core_temperature)
            + self.solar_heating.value(surface_temperature)
            - self.convective_cooling.value(surface_temperature)
            - self.radiative_cooling.value(surface_temperature)
            - self.precipitation_cooling.value(surface_temperature)
        )

    def morgan(
        self, surface_temperature: floatArray, core_temperature: floatArray
    ) -> floatArray:
        """
        Computes the Morgan function for given temperature arrays.

        Args:
            surface_temperature (numpy.ndarray): Array of surface temperatures.
            core_temperature (numpy.ndarray): Array of core temperatures.

        Returns:
            numpy.ndarray: Resulting array after applying the Morgan function.
        """
        heat_capacity, _, _, _ = self.morgan_coefficients
        return (core_temperature - surface_temperature) - heat_capacity * self.joule(
            surface_temperature, core_temperature
        ) / (2.0 * np.pi * self.args.radial_thermal_conductivity)

    def steady_temperature(
        self,
        surface_temperature_guess: Optional[floatArrayLike] = None,
        core_temperature_guess: Optional[floatArrayLike] = None,
        tol: float = default.tol,
        maxiter: int = default.maxiter,
        return_err: bool = False,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """
        Compute the steady-state temperature distribution.

        Args:
            surface_temperature_guess (float | numpy.ndarray | None): Initial guess for the surface temperature. If None, ambient temperature is used.
            core_temperature_guess (float | numpy.ndarray | None): Initial guess for the core temperature. If None, 1.5 times the absolute value of ambient temperature is used.
            tol (float): Tolerance for the quasi-Newton solver.
            maxiter (int): Maximum number of iterations for the quasi-Newton solver.
            return_err (bool): If True, the error of the solution is included in the returned DataFrame.
            return_power (bool): If True, power-related values are included in the returned DataFrame.

        Returns:
            pd.DataFrame: DataFrame containing the steady-state temperatures and optionally the error and power-related values.
        """

        # if no guess provided, use ambient temp
        shape = (self.args.get_number_of_computations(),)
        surface_temperature_guess = (
            surface_temperature_guess
            if surface_temperature_guess is not None
            else 1.0 * self.args.ambient_temperature
        )
        core_temperature_guess = (
            core_temperature_guess
            if core_temperature_guess is not None
            else 1.5 * np.abs(self.args.ambient_temperature)
        )
        surface_temperature_guess_ = surface_temperature_guess * np.ones(shape)
        core_temperature_guess_ = core_temperature_guess * np.ones(shape)

        # solve system
        x, y, iterations, err = quasi_newton_2d(
            func1=self.balance,
            func2=self.morgan,
            x_init=surface_temperature_guess_,
            y_init=core_temperature_guess_,
            relative_tolerance=tol,
            max_iterations=maxiter,
            delta_x=1.0e-03,
            delta_y=1.0e-03,
        )
        if np.max(err) > tol or iterations == maxiter:
            print(
                f"rstat_analytic max err is {np.max(err):.3E} in {iterations:d} iterations"
            )

        # format output
        z = self.average(x, y)
        df = pd.DataFrame(
            {
                TemperatureType.SURFACE: x,
                TemperatureType.AVERAGE: z,
                TemperatureType.CORE: y,
            }
        )

        if return_err:
            df[VariableType.ERROR] = err

        if return_power:
            df[PowerType.JOULE] = self.joule(x, y)
            df[PowerType.SOLAR] = self.solar_heating.value(x)
            df[PowerType.CONVECTION] = self.convective_cooling.value(x)
            df[PowerType.RADIATION] = self.radiative_cooling.value(x)
            df[PowerType.RAIN] = self.precipitation_cooling.value(x)

        return df

    def _morgan_transient(self):
        """Morgan coefficients for transient temperature."""
        heat_capacity, outer_diameter, core_diameter, ix = self.morgan_coefficients
        c1 = heat_capacity / (2.0 * np.pi * self.args.radial_thermal_conductivity)
        c2 = 0.5 * np.ones_like(c1)
        a, b = _profile_bim_avg_coeffs(
            0.5 * core_diameter[ix], 0.5 * outer_diameter[ix]
        )
        c2[ix] = a / b
        return c1, c2

    def _transient_temperature_results(
        self,
        offset,
        surface_temperature,
        average_temperature,
        core_temperature,
        return_power,
        n,
    ):
        dr = {
            VariableType.TIME: offset,
            TemperatureType.SURFACE: surface_temperature,
            TemperatureType.AVERAGE: average_temperature,
            TemperatureType.CORE: core_temperature,
        }

        if return_power:
            for power in Solver_.powers():
                dr[power] = np.zeros_like(surface_temperature)

            for i in range(len(offset)):
                dr[PowerType.JOULE][i, :] = self.joule(
                    surface_temperature[i, :], core_temperature[i, :]
                )
                dr[PowerType.SOLAR][i, :] = self.solar_heating.value(
                    surface_temperature[i, :]
                )
                dr[PowerType.CONVECTION][i, :] = self.convective_cooling.value(
                    surface_temperature[i, :]
                )
                dr[PowerType.RADIATION][i, :] = self.radiative_cooling.value(
                    surface_temperature[i, :]
                )
                dr[PowerType.RAIN][i, :] = self.precipitation_cooling.value(
                    surface_temperature[i, :]
                )

        if n == 1:
            keys = list(dr.keys())
            keys.remove(VariableType.TIME)
            for k in keys:
                dr[k] = dr[k][:, 0]

        return dr

    def transient_temperature(
        self,
        offset: floatArray = np.array([]),
        surface_temperature_0: Optional[floatArrayLike] = None,
        core_temperature_0: Optional[floatArrayLike] = None,
        return_power: bool = False,
    ) -> Dict[str, Any]:
        """
        Compute transient-state temperature.

        Args:
            offset (numpy.ndarray): A 1D array with times (in seconds) when the temperature needs to be computed. The array must contain increasing values (undefined behaviour otherwise).
            surface_temperature_0 (float | numpy.ndarray | None): Initial surface temperature. If None, the ambient temperature from the internal dict will be used. The default is None.
            core_temperature_0 (float | numpy.ndarray | None): Initial core temperature. If None, the ambient temperature from the internal dict will be used. The default is None.
            return_power (bool, optional): Return power term values. The default is False.

        Returns:
            Dict[str, Any]: A dictionary with temperature and other results (depending on inputs) in the keys.

        """
        # get sizes (n for input dict entries, N for time)
        n = self.args.get_number_of_computations()
        N = len(offset)
        if N < 2:
            raise ValueError()

        # get initial temperature
        surface_temperature_0 = (
            surface_temperature_0
            if surface_temperature_0 is not None
            else self.args.ambient_temperature
        )
        core_temperature_0 = (
            core_temperature_0
            if core_temperature_0 is not None
            else 1.0 + surface_temperature_0
        )
        time_changing_parameters = get_time_changing_parameters(self.args, offset, N, n)
        c1, c2 = self._morgan_transient()
        # inverse of m*C : shortcuts for time-loop
        imc = 1.0 / (self.args.linear_mass * self.args.heat_capacity)

        # init
        surface_temperature = np.zeros((N, n))
        average_temperature = np.zeros((N, n))
        core_temperature = np.zeros((N, n))
        surface_temperature[0, :] = surface_temperature_0
        core_temperature[0, :] = core_temperature_0
        average_temperature[0, :] = self.average(
            surface_temperature[0, :], core_temperature[0, :]
        )

        # main time loop
        for i in range(1, len(offset)):
            for k in time_changing_parameters.keys():
                self.args[k] = time_changing_parameters[k][i, :]
            self.update()
            bal = self.balance(
                surface_temperature[i - 1, :], core_temperature[i - 1, :]
            )
            average_temperature[i, :] = (
                average_temperature[i - 1, :] + (offset[i] - offset[i - 1]) * bal * imc
            )
            mrg = c1 * (self.joule_heating.value(average_temperature[i, :]) - bal)
            core_temperature[i, :] = average_temperature[i, :] + c2 * mrg
            surface_temperature[i, :] = core_temperature[i, :] - mrg

        return self._transient_temperature_results(
            offset,
            surface_temperature,
            average_temperature,
            core_temperature,
            return_power,
            n,
        )

    @staticmethod
    def _check_target(target: Optional[CableLocationListLike], core_diameter, max_len):
        """
        Validates and processes the target temperature input.

        :param target: The target temperature(s) to be validated. It can be:
            - None: which sets the target automatically.
            - A CableLocation: must be one of CableLocation.SURFACE, CableLocation.AVERAGE, or CableLocation.CORE.
            - A list of CableLocation: each element of the list must be one of CableLocation.SURFACE, CableLocation.AVERAGE, or CableLocation.CORE.
        :param core_diameter: The core diameter of the cable.
        :param max_len: The expected length of the target list if target is a list.
        :return: An array of target labels if the input is valid.
        """
        # check target
        if target is None:
            d_ = core_diameter * np.ones(max_len)
            return np.array(
                [
                    TargetType.CORE if d_[i] > 0.0 else TargetType.AVERAGE
                    for i in range(max_len)
                ]
            )
        elif isinstance(target, TargetType):
            return np.array([target] * max_len)
        return np.array(target)

    def _steady_intensity_header(
        self, T: floatArrayLike, target: CableLocationListLike
    ) -> Tuple[np.ndarray, Callable]:
        """Format input for ampacity solver."""

        max_len = self.args.get_number_of_computations()
        Tmax = T * np.ones(max_len)
        target_ = self._check_target(target, self.args.core_diameter, max_len)

        # pre-compute indexes
        heat_capacity, outer_diameter, core_diameter, ix = self.morgan_coefficients
        a, b = _profile_bim_avg_coeffs(0.5 * core_diameter, 0.5 * outer_diameter)

        js = np.nonzero(target_ == TargetType.SURFACE)[0]
        ja = np.nonzero(target_ == TargetType.AVERAGE)[0]
        jc = np.nonzero(target_ == TargetType.CORE)[0]
        jx = np.intersect1d(ix, ja)

        # get correct input for quasi-newton solver
        def newtheader(i: floatArray, tg: floatArray) -> Tuple[floatArray, floatArray]:
            self.args.transit = i
            self.joule_heating.__init__(**self.args.__dict__)
            surface_temperature = np.ones_like(tg) * np.nan
            core_temperature = np.ones_like(tg) * np.nan

            surface_temperature[js] = Tmax[js]
            core_temperature[js] = tg[js]

            surface_temperature[ja] = tg[ja]
            core_temperature[ja] = 2 * Tmax[ja] - surface_temperature[ja]
            core_temperature[jx] = (
                b[jx] * Tmax[jx] - a[jx] * surface_temperature[jx]
            ) / (b[jx] - a[jx])

            core_temperature[jc] = Tmax[jc]
            surface_temperature[jc] = tg[jc]

            return surface_temperature, core_temperature

        return Tmax, newtheader

    def steady_intensity(
        self,
        max_conductor_temperature: floatArrayLike = np.array([]),
        target: CableLocationListLike = None,
        cable_type: CableTypeListLike = None,
        tol: float = default.tol,
        maxiter: int = default.maxiter,
        return_err: bool = False,
        return_temp: bool = True,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """
        Compute the steady-state intensity for a given temperature profile.

        Args:
            max_conductor_temperature (float | numpy.ndarray): Initial temperature profile. Default is an empty numpy array.
            target (TargetType | list[CableLocation]): Target specification for the solver. Default is None.
            cable_type (CableType | list[CableType]): Cable type specification for the solver. Default is None. If provided, it overrides the target specification.
            tol (float): Tolerance for the solver. Default is DP.tol.
            maxiter (int): Maximum number of iterations for the solver. Default is DP.maxiter.
            return_err (bool): If True, return the error in the output DataFrame. Default is False.
            return_temp (bool): If True, return the temperature profiles in the output DataFrame. Default is True.
            return_power (bool): If True, return the power profiles in the output DataFrame. Default is True.

        Returns:
            pd.DataFrame: DataFrame containing the steady-state intensity and optionally the error, temperature profiles, and power profiles.
        """
        target = _infer_target_from_cable_type(cable_type, target)

        Tmax, newtheader = self._steady_intensity_header(
            max_conductor_temperature, target
        )

        def balance(i: floatArray, tg: floatArray) -> floatArrayLike:
            surface_temperature, core_temperature = newtheader(i, tg)
            return self.balance(surface_temperature, core_temperature)

        def morgan(i: floatArray, tg: floatArray) -> floatArray:
            surface_temperature, core_temperature = newtheader(i, tg)
            return self.morgan(surface_temperature, core_temperature)

        # solve system
        s = Solver1T(
            self.args.__dict__,
            type(self.joule_heating),
            type(self.solar_heating),
            type(self.convective_cooling),
            type(self.radiative_cooling),
            type(self.precipitation_cooling),
        )
        r = s.steady_intensity(Tmax, tol=1.0, maxiter=8, return_power=False)
        x, y, iterations, err = quasi_newton_2d(
            balance,
            morgan,
            r[VariableType.TRANSIT].values,
            Tmax,
            relative_tolerance=tol,
            max_iterations=maxiter,
            delta_x=1.0e-03,
            delta_y=1.0e-03,
        )
        if np.max(err) > tol or iterations == maxiter:
            print(
                f"rstat_analytic max err is {np.max(err):.3E} in {iterations:d} iterations"
            )

        # format output
        df = pd.DataFrame({VariableType.TRANSIT: x})

        if return_err:
            df[VariableType.ERROR] = err

        if return_temp or return_power:
            surface_temperature, core_temperature = newtheader(x, y)
            average_temperature = self.average(surface_temperature, core_temperature)

            if return_temp:
                df[TemperatureType.SURFACE] = surface_temperature
                df[TemperatureType.AVERAGE] = average_temperature
                df[TemperatureType.CORE] = core_temperature

            if return_power:
                df[PowerType.JOULE] = self.joule_heating.value(average_temperature)
                df[PowerType.SOLAR] = self.solar_heating.value(surface_temperature)
                df[PowerType.CONVECTION] = self.convective_cooling.value(
                    surface_temperature
                )
                df[PowerType.RADIATION] = self.radiative_cooling.value(
                    surface_temperature
                )
                df[PowerType.RAIN] = self.precipitation_cooling.value(
                    surface_temperature
                )

        return df



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Utility functions to compute different sun positions from a point on earth.

These positions usually depend on the point latitude and the time. The sun
position is then used to estimate the solar radiation in CIGRE and IEEE
models.
"""

from datetime import time
from math import pi
from typing import Iterable

import numpy as np
from thermohl import (
    floatArrayLike,
    dateListLike,
    datetimeListLike,
)


def time_to_float_hours(t: time) -> float:
    return t.hour + t.minute / 60 + t.second / 3600


def utc2solar_hour(
    datetime_utc: datetimeListLike,
    longitude: floatArrayLike,
) -> floatArrayLike:
    """Convert UTC datetime to solar hour adding the longitude contribution.
    If more than one input are numpy arrays, they should have the same size.

    :param datetime_utc: Datetime in UTC.
    :param longitude: Longitude (in rad).
    :return: Solar hour.
    """
    day_of_year = (
        np.array([d.timetuple().tm_yday for d in datetime_utc])
        if isinstance(datetime_utc, Iterable)
        else datetime_utc.timetuple().tm_yday
    )
    utc_hour = (
        np.array([time_to_float_hours(d.time()) for d in datetime_utc])
        if isinstance(datetime_utc, Iterable)
        else time_to_float_hours(datetime_utc.time())
    )
    B = 2 * pi * (day_of_year - 81) / 365
    solar_hour = (
        utc_hour
        + longitude / (2 * pi) * 24
        - (7.678 * np.sin(B + 1.374) - 9.87 * np.sin(2 * B)) / 60
    )
    return solar_hour


def hour_angle(solar_hour: floatArrayLike) -> floatArrayLike:
    """Compute hour angle.
    If more than one input are numpy arrays, they should have the same size.

    :param solar_hour: solar hour of the day.
    :return: Hour angle in radians.
    """
    return np.radians(15.0 * (solar_hour - 12.0))


def solar_declination(date: dateListLike) -> floatArrayLike:
    """Compute solar declination.
    If more than one input are numpy arrays, they should have the same size.

    :param date: Date of the year.
    :return: Solar declination in radians.
    """
    day_of_year = (
        np.array([d.timetuple().tm_yday for d in date])
        if isinstance(date, Iterable)
        else date.timetuple().tm_yday
    )
    return np.deg2rad(23.46) * np.sin(2.0 * np.pi * (day_of_year + 284) / 365.0)


def solar_altitude(
    latitude: floatArrayLike,
    date: dateListLike,
    solar_hour: floatArrayLike,
) -> floatArrayLike:
    """Compute solar altitude.
    If more than one input are numpy arrays, they should have the same size.

    :param latitude: latitude in radians.
    :param date: Date of the year.
    :param solar_hour: solar hour of the day.
    :return: Solar altitude in radians.
    """
    computed_solar_declination = solar_declination(date)
    computed_hour_angle = hour_angle(solar_hour)
    return np.arcsin(
        np.cos(latitude)
        * np.cos(computed_solar_declination)
        * np.cos(computed_hour_angle)
        + np.sin(latitude) * np.sin(computed_solar_declination)
    )


def solar_azimuth(
    latitude: floatArrayLike,
    date: dateListLike,
    solar_hour: floatArrayLike,
) -> floatArrayLike:
    """Compute solar azimuth.
    If more than one input are numpy arrays, they should have the same size.

    :param latitude: latitude in radians.
    :param date: Date of the year.
    :param solar_hour: solar hour of the day.
    :return: Solar azimuth in radians.
    """
    computed_solar_declination = solar_declination(date)
    computed_hour_angle = hour_angle(solar_hour)
    azimuth_ratio = np.sin(computed_hour_angle) / (
        np.sin(latitude) * np.cos(computed_hour_angle)
        - np.cos(latitude) * np.tan(computed_solar_declination)
    )
    azimuth_offset_rad = np.where(
        azimuth_ratio >= 0.0,
        np.where(computed_hour_angle < 0.0, 0.0, np.pi),
        np.where(computed_hour_angle < 0.0, np.pi, 2.0 * np.pi),
    )
    return azimuth_offset_rad + np.arctan(azimuth_ratio)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from thermohl import floatArrayLike


class Air:
    """Air quantities."""

    @staticmethod
    def volumic_mass(
        air_temperature: floatArrayLike, altitude: floatArrayLike = 0.0
    ) -> floatArrayLike:
        r"""Compute air volumic mass.

        If both inputs are numpy arrays, they should have the same size.

        Args:
            air_temperature (float | numpy.ndarray): Air temperature (°C).
            altitude (float | numpy.ndarray, optional): Altitude above sea level (m). The default is 0.

        Returns:
            float | numpy.ndarray: Volumic mass in kg·m⁻³.

        """
        return (1.293 - 1.525e-04 * altitude + 6.379e-09 * altitude**2) / (
            1.0 + 0.00367 * air_temperature
        )

    @staticmethod
    def dynamic_viscosity(air_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute air dynamic viscosity.

        Args:
            air_temperature (float | numpy.ndarray): Air temperature (°C)

        Returns:
            float | numpy.ndarray: Dynamic viscosity in kg·m⁻¹·s⁻¹.

        """
        return (1.458e-06 * (air_temperature + 273.0) ** 1.5) / (
            air_temperature + 383.4
        )

    @staticmethod
    def thermal_conductivity(air_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute air thermal conductivity.

        Args:
            air_temperature (float | numpy.ndarray): Air temperature (°C)

        Returns:
            float | numpy.ndarray: Thermal conductivity in W·m⁻¹·K⁻¹.

        """
        return 2.424e-02 + 7.477e-05 * air_temperature - 4.407e-09 * air_temperature**2



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

from thermohl import floatArrayLike
from thermohl.power import RadiativeCoolingBase


class RadiativeCooling(RadiativeCoolingBase):
    def __init__(
        self,
        ambient_temperature: floatArrayLike,
        outer_diameter: floatArrayLike,
        emissivity: floatArrayLike,
        stefan_boltzmann_constant: float = 5.67e-08,
        **kwargs: Any,
    ):
        r"""Init with args.

        Args:
            ambient_temperature (float | numpy.ndarray): Ambient temperature (°C).
            outer_diameter (float | numpy.ndarray): External diameter (m).
            emissivity (float | numpy.ndarray): Emissivity (—).
            stefan_boltzmann_constant (float, optional): Stefan–Boltzmann constant (W·m⁻²·K⁻⁴). The default is 5.67e-08.
        """
        super().__init__(
            ambient_temperature=ambient_temperature,
            outer_diameter=outer_diameter,
            emissivity=emissivity,
            stefan_boltzmann_constant=stefan_boltzmann_constant,
            zerok=273.0,
            **kwargs,
        )



# Copyright 2023 Eurobios Mews Labs
# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0


"""Misc. utility code for thermohl project."""

import os
from functools import wraps
from importlib.util import find_spec

import numpy as np
import pandas as pd
import yaml


def _dict_completion(
    params: dict,
    filename: str,
    validate_types: bool = True,
    warning: bool = False,
) -> dict:
    """Complete input dict with values from file.

    Read dict stored in filename (yaml format) and for each key in it, add it
    to input dict dat if the key is not already in dat.

    Args:
        params (dict): Input dict with parameters for power terms.
        warning (bool, optional): Print a message if a parameter is missing. The default is False.

    Returns:
        dict: Completed input dict if some parameters were missing.

    """
    file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    defaults = yaml.safe_load(open(file_path, "r"))
    for key in defaults.keys():
        if key not in params.keys() or params[key] is None:
            params[key] = defaults[key]
            if warning:
                print("Added key %s from default parameters" % (key,))
        elif (
            not isinstance(params[key], int)
            and not isinstance(params[key], float)
            and not isinstance(params[key], np.ndarray)
            and validate_types
        ):
            raise TypeError(
                "element in input dict (key [%s]) must be int, float or numpy.ndarray"
                % (key,)
            )
    return params


def add_default_parameters(params: dict, warning: bool = False) -> dict:
    """Add default parameters if there is missing input.

    Args:
        params (dict): Input dict with parameters for power terms.
        warning (bool, optional): Print a message if a parameter is missing. The default is False.

    Returns:
        dict: Completed input dict if some parameters were missing.

    """
    file_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "default_values.yaml"
    )
    return _dict_completion(params, file_path, warning=warning)


def add_default_uncertainties(params: dict, warning: bool = False) -> dict:
    """Add default uncertainty parameters if there is missing input.

    Args:
        params (dict): Input dict with parameters for power terms.
        warning (bool, optional): Print a message if a parameter is missing. The default is False.

    Returns:
        dict: Completed input dict if some parameters were missing.

    """
    file_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "default_uncertainties.yaml"
    )
    return _dict_completion(params, file_path, validate_types=False, warning=warning)


def df2dct(df: pd.DataFrame) -> dict:
    """Convert a pandas.DataFrame to a dictionary.

    Would be an equivalent to df.to_dict(orient='numpy.ndarray') if it existed.

    Args:
        df (pandas.DataFrame): Input DataFrame.

    Returns:
        dict: Dictionary with values converted to scalars or numpy arrays.
    """
    values_by_key = df.to_dict(orient="list")
    for key in values_by_key.keys():
        if len(values_by_key[key]) > 1:
            values_by_key[key] = np.array(values_by_key[key])
        else:
            values_by_key[key] = values_by_key[key][0]
    return values_by_key


def bisect_v(
    func: callable,
    lower_bound: float,
    upper_bound: float,
    output_shape: tuple[int, ...],
    tolerance: float = 1.0e-06,
    max_iterations: int = 128,
    print_error: bool = False,
) -> tuple[np.ndarray, np.ndarray]:
    """Bisection method to find a zero of a continuous function [a, b] -> R,
    such that f(a) <= 0 <= f(b).

    The method is vectorized, in the sense that it can in a single call find
    the zeros of several independent real-valued functions with the same input
    range [a, b].
    For this purpose, the `fun` argument should be a Python function taking as
    input a Numpy array of values in [a, b] and returning a Numpy array
    containing the evaluation of each function for the corresponding input.
    This is most efficient if the outputs values of all these functions are
    computed using vectorized Numpy operations as in the example below.

    Args:
        func (Callable[[np.ndarray], np.ndarray]): Python function taking a NumPy array and returning a NumPy array of the same shape.
        lower_bound (float): Lower bound of the [a, b] interval.
        upper_bound (float): Upper bound of the [a, b] interval.
        output_shape (tuple[int, ...]): Shape of the inputs and outputs of `func` and thus of the outputs of `bisect_v`.
        tolerance (float): Absolute tolerance.
        max_iterations (int): Maximum number of iterations.
        print_error (bool): Whether to print the max absolute error and iteration count at the end.

    Returns:
        tuple[np.ndarray, np.ndarray]:
            - x: NumPy array with the zeros (same shape as `shape`).
            - err: NumPy array with the absolute convergence error (same shape as `shape`).

    Examples
    >>> c = np.array([1.0, 4.0, 9.0, 16.0])
    >>> def f(x):
    ...     return x**2 - c
    >>> x0, err = bisect_v(f, lower_bound=0.0, upper_bound=10.0, output_shape=(4,), tolerance=1e-10)
    >>> x0
    array([1., 2., 3., 4.])

    """
    lower_bounds = lower_bound * np.ones(output_shape)
    upper_bounds = upper_bound * np.ones(output_shape)

    # If the condition f(a) <= 0 <= f(b) is not satisfied,
    # there's no guaranty there is a unique solution, so the bisection method can't be applied
    # and we raise an error
    if np.any(func(lower_bounds) > 0) or np.any(func(upper_bounds) < 0):
        raise ValueError(
            "Can't use bisection method: function applied to lower bound should be strictly negative"
            " and function applied to upper bound should be strictly positive."
        )

    abs_error = np.abs(upper_bound - lower_bound)
    iteration_count = 1
    while np.nanmax(abs_error) > tolerance and iteration_count <= max_iterations:
        midpoint = 0.5 * (lower_bounds + upper_bounds)
        values = func(midpoint)
        lower_mask = values < 0
        lower_bounds[lower_mask] = midpoint[lower_mask]
        upper_bounds[~lower_mask] = midpoint[~lower_mask]
        abs_error = np.abs(upper_bounds - lower_bounds)
        iteration_count += 1
    midpoint = 0.5 * (lower_bounds + upper_bounds)
    midpoint[np.isnan(func(midpoint))] = np.nan
    if print_error:
        print(
            f"Bisection max err (abs) : {np.max(abs_error):.2E}; count={iteration_count}"
        )
    return midpoint, abs_error


# In agreement with Eurobios, this function has been retrieved from the pyntb library,
# in order to remove the external dependency on this library.
# In this library, this function was initially developed under the name qnewt2d_v
def quasi_newton_2d(
    func1: callable,
    func2: callable,
    x_init: np.ndarray,
    y_init: np.ndarray,
    relative_tolerance: float = 1.0e-12,
    max_iterations: int = 64,
    delta_x: float = 1.0e-03,
    delta_y: float = 1.0e-03,
) -> tuple[np.ndarray, np.ndarray, int, np.ndarray]:
    """
    Two-dimensional quasi-Newton with arrays.

    Apply a 2D quasi newton on a large number of case at the same time, ie solve
    the system [f1(x, y), f2(x, y)] = [0., 0.] in n cases.

    Derivatives are estimated with a second-order centered estimation (ie f1 and
    f2 are evaluated four times at each iteration).

    All return values are arrays of the same size as inputs x0 and y0.

    Args:
        func1 (Callable[[np.ndarray, np.ndarray], np.ndarray]): First component of a 2D function of two variables.
        func2 (Callable[[np.ndarray, np.ndarray], np.ndarray]): Second component of a 2D function of two variables.
        x_init (np.ndarray): First component of the initial guess.
        y_init (np.ndarray): Second component of the initial guess.
        relative_tolerance (float): Relative tolerance for convergence.
        max_iterations (int): Maximum number of iterations.
        delta_x (float): Delta for evaluating the derivative with respect to the first component.
        delta_y (float): Delta for evaluating the derivative with respect to the second component.

    Returns:
        tuple[np.ndarray, np.ndarray, int, np.ndarray]:
            - x: First component of the solution.
            - y: Second component of the solution.
            - count: Number of iterations when exiting the function.
            - err: Relative error when exiting the function (per component).

    """
    x = x_init.copy()
    y = y_init.copy()

    for count in range(max_iterations):
        # Evaluate functions at current x and y
        func1_value = func1(x, y)
        func2_value = func2(x, y)

        # Compute Jacobian matrix using second-order centered differences
        jacobian_11 = (func1(x + delta_x, y) - func1(x - delta_x, y)) / (2 * delta_x)
        jacobian_12 = (func1(x, y + delta_y) - func1(x, y - delta_y)) / (2 * delta_y)
        jacobian_21 = (func2(x + delta_x, y) - func2(x - delta_x, y)) / (2 * delta_x)
        jacobian_22 = (func2(x, y + delta_y) - func2(x, y - delta_y)) / (2 * delta_y)

        # Compute inverse of the Jacobian determinant
        inv_jacobian_det = 1.0 / (jacobian_11 * jacobian_22 - jacobian_12 * jacobian_21)
        err_abs_x = inv_jacobian_det * (
            jacobian_22 * func1_value - jacobian_12 * func2_value
        )
        err_abs_y = inv_jacobian_det * (
            jacobian_11 * func2_value - jacobian_21 * func1_value
        )

        x -= err_abs_x
        y -= err_abs_y

        # Check for convergence
        err = max(np.nanmax(np.abs(err_abs_x / x)), np.nanmax(np.abs(err_abs_y / y)))
        if err <= relative_tolerance:
            break

    return x, y, count + 1, np.maximum(np.abs(err_abs_x / x), np.abs(err_abs_y / y))


def depends_on_optional(module_name: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            spec = find_spec(module_name)
            if spec is None:
                raise ImportError(
                    f"Optional dependency {module_name} not found ({func.__name__})."
                )
            else:
                return func(*args, **kwargs)

        return wrapper

    return decorator



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
from thermohl.power.olla import Air


def test_volumic_mass_scalar():
    air_temperature = 25.0
    altitude_meters = 0.0
    expected = 1.184

    result = Air.volumic_mass(air_temperature, altitude_meters)

    assert np.isclose(result, expected, rtol=1e-2)


def test_volumic_mass_array():
    air_temperature = np.array([0.0, 25.0, 100.0])
    altitude_meters = np.array([0.0, 1000.0, 2000.0])
    expected = np.array([1.292, 1.055, 0.787])

    result = Air.volumic_mass(air_temperature, altitude_meters)

    assert np.allclose(result, expected, rtol=1e-2)


def test_volumic_mass_altitude_scalar():
    air_temperature = 25.0
    altitude_meters = 1000.0
    expected = 1.055

    result = Air.volumic_mass(air_temperature, altitude_meters)

    assert np.isclose(result, expected, rtol=1e-2)


def test_volumic_mass_altitude_array():
    air_temperature = np.array([0.0, 25.0, 100.0])
    altitude_meters = 1000.0
    expected = np.array([1.139, 1.055, 0.862])

    result = Air.volumic_mass(air_temperature, altitude_meters)

    assert np.allclose(result, expected, rtol=1e-2)


def test_dynamic_viscosity_scalar():
    air_temperature = 25.0
    expected = 1.849e-05

    result = Air.dynamic_viscosity(air_temperature)

    assert np.isclose(result, expected, rtol=1e-2)


def test_dynamic_viscosity_array():
    air_temperature = np.array([0.0, 25.0, 100.0])
    expected = np.array([1.723e-05, 1.839e-05, 2.168e-05])

    result = Air.dynamic_viscosity(air_temperature)

    assert np.allclose(result, expected, rtol=1e-2)


def test_kinematic_viscosity_scalar():
    air_temperature = 25.0
    altitude_meters = 0.0
    expected = 1.561e-05

    result = Air.kinematic_viscosity(air_temperature, altitude_meters)

    assert np.isclose(result, expected, rtol=1e-2)


def test_kinematic_viscosity_array():
    air_temperature = np.array([0.0, 25.0, 100.0])
    altitude_meters = np.array([0.0, 1000.0, 2000.0])
    expected = np.array([1.333e-05, 1.742e-05, 2.754e-05])

    result = Air.kinematic_viscosity(air_temperature, altitude_meters)

    assert np.allclose(result, expected, rtol=1e-2)


def test_thermal_conductivity_scalar():
    air_temperature = 25.0
    expected = 0.02587

    result = Air.thermal_conductivity(air_temperature)

    assert np.isclose(result, expected, rtol=1e-2)


def test_thermal_conductivity_array():
    air_temperature = np.array([0.0, 25.0, 100.0])
    expected = np.array([0.024, 0.026, 0.0316])

    result = Air.thermal_conductivity(air_temperature)

    assert np.allclose(result, expected, rtol=1e-2)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from math import pi
from typing import Any, Tuple, Iterable
import numpy as np
from thermohl import (
    floatArrayLike,
    sun,
    datetimeListLike,
)
from thermohl.power import SolarHeatingBase


def compute_solar_irradiance(
    global_radiation: floatArrayLike,
    solar_altitude: floatArrayLike,
    incidence: floatArrayLike,
    nebulosity: floatArrayLike,
    albedo: floatArrayLike,
) -> floatArrayLike:
    """Compute solar radiation.
    Difference with IEEE version are neither turbidity or altitude influence.

    :param global_radiation: global radiation.
    :param solar_altitude: solar altitude in radians.
    :param incidence: incidence angle in radians.
    :param nebulosity: nebulosity.
    :param albedo: albedo.
    :return: Solar radiation value. Negative values are set to zero.
    """

    def compute_diffuse_radiation() -> floatArrayLike:
        return global_radiation * (0.3 + 0.7 * (nebulosity / 8) ** 2)

    def compute_beam_radiation() -> floatArrayLike:
        # si l'altitude solaire est nulle, on fixe la radiation solaire à 0.
        with np.errstate(divide="ignore", invalid="ignore"):
            return (global_radiation - diffuse_radiation) / np.sin(solar_altitude)

    diffuse_radiation = compute_diffuse_radiation()
    beam_radiation = compute_beam_radiation()
    solar_irradiance = beam_radiation * (
        np.sin(incidence) + pi / 2 * albedo * np.sin(solar_altitude)
    ) + diffuse_radiation * pi / 2 * (1 + albedo)

    return np.where(solar_altitude > 0.0, solar_irradiance, 0.0)


def compute_data_from_provided(
    provided_global_radiation: floatArrayLike,
    provided_nebulosity: floatArrayLike,
    solar_altitude: floatArrayLike,
) -> Tuple[floatArrayLike, floatArrayLike]:
    """
    Returns a value of nebulosity and a value of global_radiation.
    If the global radiation is provided (ie not NaN), the nebulosity is computed from it.
    Otherwise, the global radiation is computed from the provided nebulosity (default value of 0).

    :param provided_global_radiation: provided global radiation (W/m2).
    :param provided_nebulosity: provided nebulosity (0 to 8).
    :param solar_altitude: solar altitude in radians.
    :return: (nebulosity, global_radiation).
    """

    mask = np.isnan(provided_global_radiation)
    # First, everything is computed (np.errstate ignore warnings on NaN values)
    with np.errstate(divide="ignore", invalid="ignore"):
        # nebulosity computation
        inter_neb = np.minimum(
            1, provided_global_radiation / (910 * np.sin(solar_altitude) - 30)
        )
        computed_nebulosity = np.minimum(8, 8 * (4 / 3 * (1 - inter_neb)) ** (1 / 3.4))

        # global radiation computation
        inter_rad = 1 - 3 / 4 * (provided_nebulosity / 8) ** 3.4
        computed_global_radiation = np.maximum(
            0, (910 * np.sin(solar_altitude) - 30) * inter_rad
        )

    # Then, a filter is applied to keep useful values amongst the computations.
    final_nebulosity = np.where(mask, provided_nebulosity, computed_nebulosity)
    final_global_radiation = np.where(
        ~mask, provided_global_radiation, computed_global_radiation
    )
    return final_nebulosity, final_global_radiation


class SolarHeating(SolarHeatingBase):
    def __init__(
        self,
        latitude: floatArrayLike,
        longitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        datetime_utc: datetimeListLike,
        outer_diameter: floatArrayLike,
        solar_absorptivity: floatArrayLike,
        albedo: floatArrayLike,
        nebulosity: floatArrayLike,
        measured_solar_irradiance: floatArrayLike,
        **kwargs: Any,
    ):
        """Build with args.
        If more than one input are numpy arrays, they should have the same size.

        :param latitude: Latitude in degrees.
        :param longitude: Longitude in degrees (must be between -180 and +180 degrees).
        :param cable_azimuth: Azimuth of the conductor in degrees.
        :param datetime_utc: Datetime in UTC.
        :param outer_diameter: external diameter of the conductor.
        :param solar_absorptivity: Solar absorption coefficient of the conductor.
        :param albedo: Ground albedo.
        :param nebulosity: Sky nebulosity (0 to 8).
        :param measured_solar_irradiance: Optional measured solar irradiance (W/m2).
        """
        date = (
            [d.date() for d in datetime_utc]
            if isinstance(datetime_utc, Iterable)
            else datetime_utc.date()
        )
        solar_hour = sun.utc2solar_hour(datetime_utc, np.deg2rad(longitude))
        solar_altitude = sun.solar_altitude(np.deg2rad(latitude), date, solar_hour)
        nebulosity, global_radiation = compute_data_from_provided(
            measured_solar_irradiance, nebulosity, solar_altitude
        )
        solar_azimuth_rad = sun.solar_azimuth(np.deg2rad(latitude), date, solar_hour)
        incidence = np.arccos(
            np.cos(solar_altitude)
            * np.cos(solar_azimuth_rad - np.deg2rad(cable_azimuth))
        )

        self.solar_absorptivity = solar_absorptivity
        self.outer_diameter = outer_diameter
        self.solar_irradiance = compute_solar_irradiance(
            global_radiation,
            solar_altitude,
            incidence,
            nebulosity,
            albedo,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl.power.ieee import RadiativeCooling
from thermohl import solver


def set_default_values_scalar():
    dic = solver.default_values()
    dic["ambient_temperature"] = 40.0
    dic["outer_diameter"] = 28.14 * 1.0e-03
    dic["emissivity"] = 0.8
    return dic


def set_default_values_array():
    dic = solver.default_values()
    dic["ambient_temperature"] = np.array([25.0, 40.0])
    dic["outer_diameter"] = np.array([24.83 * 1.0e-03, 28.14 * 1.0e-03])
    dic["emissivity"] = np.array([0.9, 0.8])
    return dic


radiative_cooling_instances = [
    RadiativeCooling(**set_default_values_array()),
    RadiativeCooling(
        **set_default_values_scalar(),
    ),
]


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_radiative_cooling_value_temperature_scalar(radiative_cooling):
    temp = 100.0
    expected_result = (
        17.8
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * (
            ((temp + 273.0) / 100.0) ** 4
            - ((radiative_cooling.ambient_temp + 273.0) / 100.0) ** 4
        )
    )

    result = radiative_cooling.value(temp)

    assert np.allclose(result, expected_result, rtol=0.001)


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_radiative_cooling_value_temperature_array(radiative_cooling):
    temp = np.array([65.3, 100.0])
    expected_result = (
        17.8
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * (
            ((temp + 273.0) / 100.0) ** 4
            - ((radiative_cooling.ambient_temp + 273.0) / 100.0) ** 4
        )
    )

    result = radiative_cooling.value(temp)

    assert np.allclose(result, expected_result, rtol=0.001)


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_radiative_cooling_derivative_temperature_scalar(radiative_cooling):
    temp = 100.0
    result = radiative_cooling.derivative(temp)

    expected_result = (
        4.0
        * 1.78e-07
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * temp**3
    )

    assert np.allclose(result, expected_result, rtol=0.001)


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_radiative_cooling_derivative_temperature_array(radiative_cooling):
    temp = np.array([65.3, 100.0])
    result = radiative_cooling.derivative(temp)

    expected_result = (
        4.0
        * 1.78e-07
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * temp**3
    )

    assert np.allclose(result, expected_result, rtol=0.001)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any, Iterable

import numpy as np

from thermohl import floatArrayLike, sun as sun, datetimeListLike
from thermohl.power import PowerTerm
from thermohl.sun import time_to_float_hours


class SolarHeating(PowerTerm):
    """Solar heating term."""

    @staticmethod
    def _solar_radiation(
        latitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        albedo: floatArrayLike,
        datetime_utc: datetimeListLike,
    ) -> floatArrayLike:
        """Compute solar radiation.

        :param latitude: Latitude in radians.
        :param cable_azimuth: Azimuth of the conductor in radians.
        :param albedo: Albedo.
        :param datetime_utc: Datetime in UTC.
        :return: Solar radiation.
        """
        date = (
            np.array([d.date() for d in datetime_utc])
            if isinstance(datetime_utc, Iterable)
            else datetime_utc.date()
        )
        hour = (
            np.array([time_to_float_hours(d.time()) for d in datetime_utc])
            if isinstance(datetime_utc, Iterable)
            else time_to_float_hours(datetime_utc.time())
        )
        solar_declination_rad = sun.solar_declination(date)
        hour_angle_rad = sun.hour_angle(hour)
        solar_altitude_rad = sun.solar_altitude(latitude, date, hour)
        direct_irradiance = (
            1280.0 * np.sin(solar_altitude_rad) / (0.314 + np.sin(solar_altitude_rad))
        )
        solar_azimuth_rad = np.arcsin(
            np.cos(solar_declination_rad)
            * np.sin(hour_angle_rad)
            / np.cos(solar_altitude_rad)
        )
        incidence_angle_rad = np.arccos(
            np.cos(solar_altitude_rad) * np.cos(solar_azimuth_rad - cable_azimuth)
        )
        direct_term = 0.5 * np.pi * albedo * np.sin(solar_altitude_rad) + np.sin(
            incidence_angle_rad
        )
        sin_altitude = np.sin(solar_altitude_rad)
        clear_sky_factor = np.piecewise(
            sin_altitude,
            [sin_altitude < 0.0, sin_altitude >= 0.0],
            [lambda value: 0.0, lambda value: value**1.2],
        )
        diffuse_term = (
            0.5
            * np.pi
            * (1 + albedo)
            * (570.0 - 0.47 * direct_irradiance)
            * clear_sky_factor
        )
        return np.where(
            solar_altitude_rad > 0.0,
            direct_term * direct_irradiance + diffuse_term,
            0.0,
        )

    def __init__(
        self,
        latitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        albedo: floatArrayLike,
        datetime_utc: datetimeListLike,
        outer_diameter: floatArrayLike,
        solar_absorptivity: floatArrayLike,
        measured_solar_irradiance: floatArrayLike,
        **kwargs: Any,
    ):
        """Init with args.
        If more than one input are numpy arrays, they should have the same size.

        :param latitude: Latitude in degrees.
        :param cable_azimuth: Azimuth of the conductor in degrees.
        :param albedo: Albedo.
        :param datetime_utc: Datetime in UTC.
        :param outer_diameter: external diameter of the conductor.
        :param solar_absorptivity: Solar absorption coefficient of the conductor.
        :param measured_solar_irradiance: Optional precomputed solar radiation term.
        """
        self.solar_absorptivity = solar_absorptivity

        mask = np.isnan(measured_solar_irradiance)
        self.solar_irradiance = np.empty_like(measured_solar_irradiance)
        if np.any(~mask):
            self.solar_irradiance[~mask] = np.maximum(measured_solar_irradiance, 0.0)
        if np.any(mask):
            self.solar_irradiance[mask] = SolarHeating._solar_radiation(
                np.deg2rad(latitude),
                np.deg2rad(cable_azimuth),
                albedo,
                datetime_utc,
            )

        self.outer_diameter = outer_diameter

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        """Compute solar heating.
        If more than one input are numpy arrays, they should have the same size.

        :param conductor_temperature: Conductor temperature (°C).
        :return: Power term value (W·m⁻¹).
        """
        return (
            self.solar_absorptivity
            * self.solar_irradiance
            * self.outer_diameter
            * np.ones_like(conductor_temperature)
        )

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        """Compute solar heating derivative."""
        return np.zeros_like(conductor_temperature)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from enum import Enum
from typing import Union, List
import numpy as np


class TargetType(Enum):
    """
    Defines the locations in the cable of the reference temperature
    * SURFACE: the reference temperature is at the surface of the cable
    * AVERAGE: the reference temperature is at an average thickness of the cable
    * CORE: the reference temperature is at the core of the cable
    """

    SURFACE = "surface"
    AVERAGE = "average"
    CORE = "core"


CableLocationListLike = Union[TargetType, List[TargetType]]


class CableType(Enum):
    """
    Defines the type of the cable.
    """

    HOMOGENEOUS = "homogeneous"
    BIMETALLIC = "bimetallic"


CableTypeListLike = Union[CableType, np.ndarray[CableType]]


class HeatEquationType(Enum):
    """
    Defines all the possible heat equation types and their string values.
    * WITH_ONE_TEMPERATURE: computes a single temperature for the cable
    * WITH_THREE_TEMPERATURES: computes three temperatures for the cable (core, surface and average)
    * WITH_THREE_TEMPERATURES_LEGACY: computes three temperatures for the cable (core, surface and average), with specifications
    """

    ONE_TEMPERATURE = "1t"
    THREE_TEMPERATURES = "3t"
    THREE_TEMPERATURES_LEGACY = "3tl"


class PowerType(Enum):
    """
    All the powers involved in the cable heating and cooling.
    * Joule heating : the way the electricity heats the cable (+ magnetic effect for core-metal cables)
    * Solar heating : the way the sun heats the cable
    * Convective cooling : the way the air cools the cable
    * Radiative cooling : the way the cable is cooled down by the radiations it emits
    * Precipitation : the way the rain cools the cable
    """

    JOULE = "joule_power"
    SOLAR = "solar_power"
    CONVECTION = "convection_power"
    RADIATION = "radiation_power"
    RAIN = "precipitation_power"


class ModelType(Enum):
    """
    All the models available in thermohl.
    * cigre : model published by CIGRE
    * ieee : model published by IEEE
    * olla : model developed by RTE's R&D team
    * rte : model developed by RTE
    """

    CIGRE = "cigre"
    IEEE = "ieee"
    OLLA = "olla"
    RTE = "rte"


class TemperatureType(Enum):
    """
    Defines all the possible temperature locations for the cable.
    * SURFACE: the temperature at the surface of the cable
    * AVERAGE: the average temperature of the cable
    * CORE: the temperature at the core of the cable
    """

    SURFACE = "surface_temperature"
    AVERAGE = "average_temperature"
    CORE = "core_temperature"


class VariableType(Enum):
    """
    Defines the different types of variables that can be used in a solver.
    * ERROR: used to retrieve the error of the solver
    * TIME: the moment associated with TRANSIT and TEMPERATURE predictions
    * TRANSIT: the electric transit used in the solver
    * TEMPERATURE: the temperature used in the solver
    """

    ERROR = "error"
    TIME = "time"
    TRANSIT = "transit"
    TEMPERATURE = "temperature"



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl.power.radiative_cooling import RadiativeCoolingBase

radiative_cooling_instances = [
    RadiativeCoolingBase(
        ambient_temperature=np.array([25.0, 15.2]),
        outer_diameter=np.array([1, 1.5]),
        emissivity=np.array([0.9, 0.8]),
    ),
    RadiativeCoolingBase(ambient_temperature=25, outer_diameter=1, emissivity=0.9),
]


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_celsius2kelvin_scalar(radiative_cooling):
    assert radiative_cooling._celsius2kelvin(0) == 273.15
    assert radiative_cooling._celsius2kelvin(100) == 373.15


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_celsius2kelvin_array(radiative_cooling):
    temp_celsius = np.array([0, 100, -273.15])

    temp_kelvin = radiative_cooling._celsius2kelvin(temp_celsius)

    np.testing.assert_array_equal(temp_kelvin, np.array([273.15, 373.15, 0]))


def test_celsius2kelvin_with_custom_zerok():
    radiative_cooling = RadiativeCoolingBase(
        ambient_temperature=25, outer_diameter=1, emissivity=0.9, zerok=0
    )

    assert radiative_cooling._celsius2kelvin(0) == 0
    assert radiative_cooling._celsius2kelvin(100) == 100


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_value_scalar(radiative_cooling):
    conductor_temp = 100

    expected_value = (
        np.pi
        * radiative_cooling.stefan_boltzmann
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * (
            (radiative_cooling._celsius2kelvin(conductor_temp)) ** 4
            - radiative_cooling.ambient_temperature**4
        )
    )

    assert np.allclose(expected_value, radiative_cooling.value(conductor_temp))


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_value_array(radiative_cooling):
    conductor_temp = np.array([50, 100])

    expected_value = (
        np.pi
        * radiative_cooling.stefan_boltzmann
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * (
            (radiative_cooling._celsius2kelvin(conductor_temp)) ** 4
            - radiative_cooling.ambient_temperature**4
        )
    )

    np.testing.assert_array_almost_equal(
        radiative_cooling.value(conductor_temp), expected_value
    )


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_derivative_scalar(radiative_cooling):
    conductor_temperature = 100

    expected = (
        4.0
        * np.pi
        * radiative_cooling.stefan_boltzmann
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * conductor_temperature**3
    )
    assert np.allclose(expected, radiative_cooling.derivative(conductor_temperature))


@pytest.mark.parametrize(
    "radiative_cooling",
    radiative_cooling_instances,
    ids=[
        "RadiativeCooling with arrays",
        "RadiativeCooling with scalars",
    ],
)
def test_derivative_array(radiative_cooling):
    # radiative_cooling = RadiativeCoolingBase(ambient_temperature=25, outer_diameter=1, emissivity=0.9, zerok=0)
    conductor_temperature = np.array([100, 200])

    expected = (
        4.0
        * np.pi
        * radiative_cooling.stefan_boltzmann
        * radiative_cooling.emissivity
        * radiative_cooling.outer_diameter
        * conductor_temperature**3
    )
    np.testing.assert_array_equal(
        radiative_cooling.derivative(conductor_temperature), expected
    )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
import numpy as np
import pytest

from thermohl.power.cigre import SolarHeating


def test_solar_radiation_scalar():
    latitude = 45.0
    cable_azimuth = 180.0
    albedo = 0.2
    datetime_utc = [datetime(2000, 6, 21, 12, tzinfo=timezone.utc)]
    expected = 1309.2

    result = SolarHeating._solar_radiation(
        latitude, cable_azimuth, albedo, datetime_utc
    )

    assert np.allclose(
        result, expected, atol=1e-1
    ), f"Expected {expected}, but got {result}"


def test_solar_radiation_array():
    latitude = np.array([45.0, 50.0, 55.0])
    cable_azimuth = np.array([180.0, 180.0, 180.0])
    albedo = np.array([0.2, 0.2, 0.2])
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
    ]
    expected = np.array([1309.2, 1267.965, 0.0])

    result = SolarHeating._solar_radiation(
        latitude, cable_azimuth, albedo, datetime_utc
    )

    assert np.allclose(
        result, expected, atol=1e-1
    ), f"Expected {expected}, but got {result}"


def test_solar_radiation_default_albedo():
    latitude = np.array([45.0, 50.0, 55.0])
    cable_azimuth = np.array([180.0, 180.0, 180.0])
    albedo = 0.2
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
    ]
    expected = np.array([1309.2, 1267.965, 0.0])

    result = SolarHeating._solar_radiation(
        latitude, cable_azimuth, albedo, datetime_utc
    )

    assert np.allclose(
        result, expected, atol=1e-1
    ), f"Expected {expected}, but got {result}"


def test_solar_radiation_mismatched_array_sizes():
    latitude = np.array([45.0, 50.0])
    cable_azimuth = np.array([180.0, 180.0, 180.0])
    albedo = np.array([0.2, 0.2])
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
    ]
    with pytest.raises(ValueError):
        SolarHeating._solar_radiation(latitude, cable_azimuth, albedo, datetime_utc)


solar_heating_instances = [
    (
        SolarHeating(
            latitude=np.array([45.0, 50.0, 55.0]),
            cable_azimuth=np.array([180.0, 180.0, 180.0]),
            albedo=np.array([0.2, 0.2, 0.2]),
            datetime_utc=[
                datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
                datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
                datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
            ],
            outer_diameter=np.array([0.01, 0.01, 0.01]),
            solar_absorptivity=np.array([0.9, 0.9, 0.9]),
            measured_solar_irradiance=np.array([np.nan, np.nan, np.nan]),
        ),
        np.array([12.3978, 11.8759, 11.2545]),
    ),
    (
        SolarHeating(
            latitude=45.0,
            cable_azimuth=180.0,
            albedo=0.2,
            datetime_utc=datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
            outer_diameter=0.01,
            solar_absorptivity=0.9,
            measured_solar_irradiance=np.array([np.nan, np.nan, np.nan]),
        ),
        12.3978,
    ),
]


@pytest.mark.parametrize(
    "solar_heating, expected",
    solar_heating_instances,
    ids=[
        "SolarHeating with arrays",
        "SolarHeating with scalars",
    ],
)
def test_solar_heating_value_scalar(solar_heating, expected):
    conductor_temperature = 25.0

    result = solar_heating.value(conductor_temperature)

    assert np.allclose(
        result, expected, atol=1e-3
    ), f"Expected {expected}, but got {result}"


def test_solar_heating_value_array():
    latitude = np.array([45.0, 50.0, 55.0])
    cable_azimuth = np.array([180.0, 180.0, 180.0])
    albedo = np.array([0.2, 0.2, 0.2])
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
    ]
    outer_diameter = np.array([0.01, 0.01, 0.01])
    solar_absorptivity = np.array([0.9, 0.9, 0.9])
    conductor_temperature = np.array([25.0, 30.0, 35.0])
    solar_heating = SolarHeating(
        latitude,
        cable_azimuth,
        albedo,
        datetime_utc,
        outer_diameter,
        solar_absorptivity,
        np.array([np.nan, np.nan, np.nan]),
    )
    expected = np.array([12.3978, 11.8759, 11.2545])

    result = solar_heating.value(conductor_temperature)

    assert np.allclose(
        result, expected, atol=1e-3
    ), f"Expected {expected}, but got {result}"


def test_solar_heating_value_mismatched_array_sizes_should_raise_error():
    latitude = np.array([45.0, 50.0])
    cable_azimuth = np.array([180.0, 180.0, 180.0])
    albedo = np.array([0.2, 0.2])
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
    ]
    outer_diameter = np.array([0.01, 0.01])
    solar_absorptivity = np.array([0.9, 0.9])
    conductor_temperature = np.array([25.0, 30.0])
    with pytest.raises(ValueError):
        solar_heating = SolarHeating(
            latitude,
            cable_azimuth,
            albedo,
            datetime_utc,
            outer_diameter,
            solar_absorptivity,
            np.array([np.nan, np.nan]),
        )
        solar_heating.value(conductor_temperature)


@pytest.mark.parametrize(
    "solar_heating, expected",
    solar_heating_instances,
    ids=[
        "SolarHeating with arrays",
        "SolarHeating with scalars",
    ],
)
def test_solar_heating_derivative_temperature_scalar(solar_heating, expected):
    conductor_temperature = 25.0
    expected = np.zeros_like(conductor_temperature)

    result = solar_heating.derivative(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "solar_heating, expected",
    solar_heating_instances,
    ids=[
        "SolarHeating with arrays",
        "SolarHeating with scalars",
    ],
)
def test_solar_heating_derivative_temperature_array(solar_heating, expected):
    conductor_temperature = np.array([25.0, 30.0, 35.0])
    expected = np.zeros_like(conductor_temperature)

    result = solar_heating.derivative(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_solar_heating_derivative_mismatched_array_sizes_should_raise_error():
    latitude = np.array([45.0, 50.0])
    cable_azimuth = np.array([180.0, 180.0, 180.0])
    albedo = np.array([0.2, 0.2])
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
    ]
    outer_diameter = np.array([0.01, 0.01])
    solar_absorptivity = np.array([0.9, 0.9])
    conductor_temperature = np.array([25.0, 30.0])
    with pytest.raises(ValueError):
        solar_heating = SolarHeating(
            latitude,
            cable_azimuth,
            albedo,
            datetime_utc,
            outer_diameter,
            solar_absorptivity,
            np.array([np.nan, np.nan]),
        )
        solar_heating.derivative(conductor_temperature)
